import collections
import dataclasses
import itertools
import logging
import time
import typing

import git
import git.exc as gitexc
import github3
import github3.exceptions as gh3e
import github3.pulls as gh3p
import github3.structs as gh3s
import yaml
import yaml.scanner

import ccc
import ccc.github
import ocm
import github.util
import gitutil
import release_notes.model as rnm

_meta_key = 'gardener.cloud/release-notes-metadata/v1'
logger = logging.getLogger(__name__)


# pylint: disable=protected-access
# noinspection PyProtectedMember
def list_associated_pulls(
        gh: github3.GitHub,
        owner: str,
        repo: str,
        sha: str
) -> typing.Optional[tuple[gh3p.ShortPullRequest]]:
    ''' Returns a tuple with pull requests related to the specified commit.

    :param gh: Instance of the GitHub v3 API
    :param owner: Owner of the repository (on GitHub)
    :param repo: Name of the repository (on GitHub)
    :param sha: SHA of the commit
    :return: a tuple with pull requests related to the specific commit
    '''
    try:
        url = gh._build_url('repos', owner, repo, 'commits', sha, 'pulls')
        return tuple(gh._iter(-1, url, gh3p.ShortPullRequest))
    except gh3e.UnprocessableEntity as e:
        logger.debug(f'cannot find any pull request related to commit {sha}: {e}')
        return None


# pylint: disable=protected-access
# noinspection PyProtectedMember
def list_pulls(
        gh: github3.GitHub,
        owner: str,
        repo: str,
        state: str = 'closed'
) -> gh3s.GitHubIterator[gh3p.ShortPullRequest]:
    url = f'{gh._build_url("repos", owner, repo, "pulls")}?state={state}'
    return gh._iter(-1, url, gh3p.ShortPullRequest)


def shorten(
        message: str,
        max_len: int = 128
) -> str:
    message = message.replace('\n', '\\n')
    if len(message) > max_len:
        message = f'{message[:max_len - 3]}...'
    return message


def create_release_notes_blocks(
        release_notes: set[rnm.ReleaseNote]
) -> str:
    return '\n\n'.join(z.block_str for z in release_notes)


def _find_git_notes_for_commit(
        repo: git.Repo,
        commit: git.Commit
) -> typing.Optional[str]:
    try:
        return repo.git.notes('show', commit.hexsha)
    except gitexc.GitCommandError as e:
        logger.debug(f'commit {commit.hexsha} does not have a git note: {e}')
        return None


def _normalize_dict_keys(
        dic: dict,
        recursive: bool = False
) -> dict:
    return {
        k.replace('-', '_').replace(' ', '_'):
        _normalize_dict_keys(v) if recursive and isinstance(v, dict) else v
        for k, v in dic.items()
    }


def _is_meta_document(doc) -> bool:
    return 'meta' in doc and isinstance(doc['meta'], dict) \
        and 'type' in doc['meta'] and isinstance(doc['meta']['type'], str) \
        and 'data' in doc['meta'] and isinstance(doc['meta']['data'], dict)


def _find_first_document(
        documents: list,
        key: str,
        ctor
):
    for doc in documents:
        if not _is_meta_document(doc):
            continue
        if doc['meta']['type'] != key:
            continue
        return ctor(**doc['meta']['data'])
    return None


def _upsert_document(
        documents: list,
        type_key: str,
        instance
) -> None:
    ''' The function searches for a (meta-) document in the given list of
    documents with the given type.  If a document was found, it updates the
    document with the given instance, otherwise it appends the instance to the
    list.

    :param documents: a (mutable) list of dicts
    :param type_key: the type to search for in the list of dicts
    :param instance: the object to insert/update in documents
    '''
    index = None
    for i, doc in enumerate(documents):
        if _is_meta_document(doc) and doc['meta']['type'] == type_key:
            index = i
            break
    if index is not None:
        documents[index] = instance
    else:
        documents.append(instance)


# Taken from the itertools recipes
# (https://docs.python.org/3/library/itertools.html#itertools-recipes)
def _grouper(iterable, n, *, incomplete='fill', fillvalue=None):
    "Collect data into non-overlapping fixed-length chunks or blocks"
    args = [iter(iterable)] * n
    if incomplete == 'fill':
        return itertools.zip_longest(*args, fillvalue=fillvalue)
    if incomplete == 'strict':
        return zip(*args, strict=True)
    if incomplete == 'ignore':
        return zip(*args)
    else:
        raise ValueError('Expected fill, strict, or ignore')


def request_pull_requests_from_api(
        git_helper: gitutil.GitHelper,
        gh: github3.GitHub,
        owner: str,
        repo_name: str,
        commits: list[git.Commit],
        group_size: int = 200,
        min_seconds_per_group: int = 300,
) -> dict[str, list[gh3p.ShortPullRequest]]:
    ''' This function requests pull requests from the GitHub API and returns a
    dictionary mapping commit SHA to a list of pull requests.

    We use notes to store the associated pull request numbers to reduce
    requests to GitHub (rate limiting).  The corresponding pull request number
    is stored in a note.  We can then fetch a list of pull requests for a
    repository and thus (theoretically) process 100 pull requests with one API
    call in the best case.

    If there is no note, request the "normal" API route to retrieve associated
    pull requests and store the pull-numbers in the commit note.
    '''
    # pr_number -> [ list of commit sha ]
    pending = collections.defaultdict(list)
    # commit_sha -> [ list of pull requests ]
    result = collections.defaultdict(list)

    # used to avoid waiting for the last group
    break_early = False

    for commit_group in _grouper(commits, group_size):
        start_time = time.time()
        for commit in commit_group:
            if commit is None:
                # groups shorter than group_size are filled with None - we can safely
                # break if we encounter one.
                break_early = True
                break

            yaml_documents = []
            is_yaml_content = True
            if note_content := _find_git_notes_for_commit(git_helper.repo, commit):
                try:
                    yaml_documents = list(yaml.safe_load_all(note_content))
                except yaml.scanner.ScannerError as e:  # YAML parsing error
                    logger.debug(
                        f'the notes of commit {commit.hexsha} do not contain valid YAML: {e}'
                    )
                    is_yaml_content = False

            # if there is already a ReleaseNotesMetadata
            if nums_meta := _find_first_document(
                documents=yaml_documents,
                key=_meta_key,
                ctor=rnm.ReleaseNotesMetadata,
            ):
                for num in nums_meta.prs:
                    pending[num].append(commit.hexsha)
                continue

            if prs := list_associated_pulls(gh, owner, repo_name, commit.hexsha):
                # add all found pull requests to the result right away
                result[commit.hexsha].extend(prs)
                # only write notes to commit if there are no notes yet,
                # or if the notes are in the YAML format already
                if note_content or not is_yaml_content:
                    continue
                data = dataclasses.asdict(
                    rnm.ReleaseNotesMetadata(round(time.time() * 1000), [z.number for z in prs])
                )
                meta = rnm.get_meta_obj(_meta_key, data)
                _upsert_document(yaml_documents, _meta_key, meta)
                git_helper.add_note(body=yaml.safe_dump_all(yaml_documents), commit=commit)

        end_time = time.time()
        time_elapsed = end_time - start_time # in seconds
        if not break_early and time_elapsed < min_seconds_per_group:
            wait_period = min_seconds_per_group - time_elapsed
            logger.info(
                f'Processed {group_size} commits in {int(time_elapsed)} seconds, will '
                f'wait {int(wait_period)} seconds before continuing.'
            )
            time.sleep(wait_period)
        # make sure to always use github-user with largest remaining quota
        gh = ccc.github.github_api(github_cfg=git_helper.github_cfg)

    if pending:
        for pull in list_pulls(gh, owner, repo_name):
            if pull.number in pending:
                for sha in pending[pull.number]:
                    result[sha].append(pull)
                del pending[pull.number]
            if len(pending) == 0:
                break
        else:
            logger.warning('one or more associated pull requests for the commits ' +
                           f'{pending.keys()} is/are either not closed or cannot be found')

    return result


def github_helper_from_github_access(
    github_access=ocm.GithubAccess,
):
    logger.info(f'Creating GH Repo-helper for {github_access.repoUrl}')
    return github.util.GitHubRepositoryHelper(
        github_api=ccc.github.github_api_from_gh_access(github_access),
        owner=github_access.org_name(),
        name=github_access.repository_name(),
    )


def git_helper_from_github_access(
    github_access: ocm.GithubAccess,
    repo_path: str,
):
    logger.info(f'Creating Git-helper for {github_access.repoUrl}')
    return gitutil.GitHelper(
        repo=repo_path,
        github_cfg=ccc.github.github_cfg_for_repo_url(github_access.repoUrl),
        github_repo_path=f'{github_access.org_name()}/{github_access.repository_name()}',
    )



import unittest

from concourse.model.resources import ResourceIdentifier


class ResourceIdentifierTest(unittest.TestCase):
    def test_ctor_arg_validation(self):
        examinee = ResourceIdentifier

        # valid invocations
        examinee(
            type_name='foo',
            base_name='bar',
            branch_name='baz',
            qualifier='qual',
            logical_name='x',
        )
        examinee(type_name='foo', base_name='bar', branch_name='baz', qualifier='qual')
        examinee(type_name='foo', base_name='bar', branch_name='baz',)

        with self.assertRaises(ValueError):
            examinee(type_name=None, base_name='bar', branch_name='baz')

        with self.assertRaises(ValueError):
            examinee(type_name='foo', base_name=None, branch_name='baz')

        with self.assertRaises(ValueError):
            examinee(type_name='foo', base_name='bar', branch_name=None)

    def test_ctor(self):
        examinee = ResourceIdentifier
        result = examinee(
            type_name='foo',
            base_name='bar',
            branch_name='baz',
            qualifier='qual',
            logical_name='x'
        )

        self.assertEqual(result.type_name(), 'foo')
        self.assertEqual(result.base_name(), 'bar')
        self.assertEqual(result.branch_name(), 'baz')
        self.assertEqual(result.logical_name(), 'x')
        self.assertEqual(result.name(), 'foo-bar-baz-qual')

        result = examinee(type_name='foo', base_name='bar', branch_name='baz', qualifier=None)

        self.assertEqual(result.name(), 'foo-bar-baz')

    def test_equal(self):
        examinee = ResourceIdentifier

        left = examinee(
            type_name='type1', base_name='base1', branch_name='branch1', qualifier='qual1',
            )
        right = examinee(
            type_name='type1', base_name='base1', branch_name='branch1', qualifier='qual1',
            )

        self.assertEqual(left, right)

        # differs in type_name
        self.assertNotEqual(
            left,
            examinee(
                type_name='type2',
                base_name='base1',
                branch_name='branch1',
                qualifier='qual1',
            ),
        )
        # differs in base_name
        self.assertNotEqual(
            left,
            examinee(
                type_name='type1',
                base_name='base2',
                branch_name='branch1',
                qualifier='qual1',
            ),
        )
        # differs in branch_name
        self.assertNotEqual(
            left,
            examinee(
                type_name='type1',
                base_name='base1',
                branch_name='branch2',
                qualifier='qual1',
            ),
        )
        # differs in qualifier
        self.assertNotEqual(
            left,
            examinee(
                type_name='type1',
                base_name='base1',
                branch_name='branch1',
                qualifier='qual2',
            ),
        )
        # absent qualifier
        self.assertNotEqual(
            left,
            examinee(
                type_name='type1',
                base_name='base1',
                branch_name='branch1',
            ),
        )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import dataclasses
from enum import Enum
import typing
import re
import urllib.parse

import dacite

import ci.util
import reutil

from . import cluster_domain_from_kubernetes_config

from model.base import (
    BasicCredentials,
    NamedModelElement,
    ModelValidationError,
    ModelBase,
)


class ConcourseOAuthConfig(NamedModelElement):
    def client_id(self) -> str:
        return self.raw['client_id']

    def client_secret(self) -> str:
        return self.raw['client_secret']

    def _required_attributes(self):
        return [
            'client_id',
            'client_secret'
        ]


@dataclasses.dataclass
class TeamWorkersCfg:
    '''
    Allows to configure dedicated workers per team, per platform.

    Useful to segregate worker nodes of critical jobs.
    This provides additional isolation on worker node level in case of worker node escape.
    '''
    team_name: str # must match concourse team name
    platform_name: str # must match a platform
    worker: int # count of workers


@dataclasses.dataclass
class Platform:
    '''
    a slightly opinionated platform name, describing a combination of an operating system and
    hardware (/CPU) architecture.

    name: arbitrary, chosen name to reference the platform (agnostic of 3rd-parties, should not
          be exposed externally)
    oci_name: the name this platform is named in the context of OCI/Docker
    worker_tag: the tag (concourse) workers of this platform are marked with

    all values are _not_ intended for being parsed/interpreted (except by e.g. containerd when being
    passed oci_name).

    --
    see https://github.com/containerd/containerd/blob/v1.5.7/platforms/platforms.go#L63 for
    allowed/known platform names

    different than stated there, oci_names *must* always specify os and  architecture here.
    '''
    name: str
    oci_name: str
    worker_tag: str | None

    def matches_oci_platform_name(self, oci_platform_name: str):
        return Platform.normalise_oci_platform_name(self.oci_name) == \
            Platform.normalise_oci_platform_name(oci_platform_name)

    @property
    def normalised_oci_platform_name(self):
        return Platform.normalise_oci_platform_name(self.oci_name)

    @property
    def normalised_oci_platform_tag_suffix(self):
        return self.normalised_oci_platform_name.replace('/', '-')

    @staticmethod
    def normalise_oci_platform_name(name: str):
        osname, arch = name.split('/', 1) # we want to fail if no / present

        if arch == 'aarch64':
            arch = 'arm64'
        elif arch == 'armhf':
            arch = 'arm'
        elif arch == 'armel':
            arch = 'arm/v6'
        elif arch == 'i386':
            arch = '386'
        elif arch in ('x86_64', 'x86-64'):
            arch = 'amd64'

        return f'{osname}/{arch}'


@dataclasses.dataclass
class WorkerNodeConfig:
    '''
    concourse worker node configuration. For now, configuration only contains platform (os/arch).

    if not configured, all nodes are assumed to run on the same, implicit default platform
    '''
    default_platform_name: str = None
    platforms: typing.List[Platform] | None = None
    team_workers_cfgs: typing.List[TeamWorkersCfg] | None = None

    def platform_for_oci_platform(self, oci_platform_name: str, absent_ok=True) -> Platform | None:
        if absent_ok and not self.platforms:
            return None

        for platform in self.platforms:
            if platform.matches_oci_platform_name(oci_platform_name):
                return platform
        if absent_ok:
            return None
        raise ValueError(f'no platform for {oci_platform_name=} found')

    def platform_for_name(self, platform_name: str, absent_ok=True) -> Platform | None:
        if absent_ok and not self.platforms:
            return None

        for platform in self.platforms:
            if platform.name == platform_name:
                return platform
        if absent_ok:
            return None
        raise ValueError(f'no platform for {platform_name=} found')

    @property
    def default_platform(self):
        if not self.default_platform_name:
            return None

        for platform in self.platforms:
            if platform.name == self.default_platform_name:
                return platform

        raise ValueError('default platform name was configured, but not found in platforms')


class ConcourseConfig(NamedModelElement):
    '''
    Not intended to be instantiated by users of this module
    '''

    def external_url(self):
        return self.raw.get('externalUrl')

    def job_mapping_cfg_name(self):
        return self.raw.get('job_mapping')

    def helm_chart_default_values_config(self):
        return self.raw.get('helm_chart_default_values_config')

    def helm_chart_values(self):
        return self.raw.get('helm_chart_values', None)

    def image_pull_secret(self):
        return self.raw.get('imagePullSecret')

    def tls_secret_name(self):
        return self.raw.get('tls_secret_name')

    def kubernetes_cluster_config(self):
        return self.raw.get('kubernetes_cluster_config')

    def ingress_host(self, config_factory):
        cluster_domain = cluster_domain_from_kubernetes_config(
            config_factory,
            self.kubernetes_cluster_config(),
        )
        return f'concourse.{cluster_domain}'

    def ingress_config(self):
        return self.raw.get('ingress_config')

    def ingress_url(self, config_factory):
        return 'https://' + self.ingress_host(config_factory)

    def helm_chart_version(self):
        return self.raw.get('helm_chart_version')

    def github_enterprise_host(self):
        return self.raw.get('github_enterprise_host')

    def proxy(self):
        return self.raw.get('proxy')

    def deploy_storage_class(self):
        return self.raw.get('deploy_storage_class', False)

    def oauth_config_name(self):
        return self.raw['oauth_config_name']

    def concourse_endpoint_name(self):
        return self.raw.get('concourse_endpoint_name')

    @property
    def worker_node_cfg(self) -> WorkerNodeConfig:
        return dacite.from_dict(
            data_class=WorkerNodeConfig,
            data=self.raw.get('worker_node_cfg', {}),
        )

    def is_accessible_from(self, url: str) -> bool:
        if not (domain_rules := self.raw.get('domain_rules', [])):
            return True

        for dr in domain_rules:
            if (
                re.compile(dr['domain_regex']).fullmatch(url)
                and not dr['accessible_from']
            ):
                return False

        return True

    def _required_attributes(self):
        return [
            'externalUrl',
            'helm_chart_default_values_config',
            'kubernetes_cluster_config',
            'job_mapping',
            'imagePullSecret',
            'tls_secret_name',
            'ingress_config',
            'helm_chart_version',
            'helm_chart_values',
            'oauth_config_name',
            'concourse_endpoint_name',
        ]

    def _optional_attributes(self):
        return {
            'concourse_version', # TODO: Remove
            'deploy_storage_class',
            'domain_rules',
            'github_enterprise_host',
            'ingress_host', # TODO: Remove
            'proxy',
            'worker_node_cfg',
        }

    def validate(self):
        super().validate()


class ConcourseEndpoint(NamedModelElement):
    def base_url(self) -> str:
        return self.raw.get('base_url')


class ConcourseTeamConfig(NamedModelElement):
    def concourse_endpoint_name(self) -> str:
        return self.raw['concourse_endpoint_name']

    def service_user(self) -> typing.Optional[BasicCredentials]:
        if local_user := self.raw.get('service_user'):
            return BasicCredentials(raw_dict=local_user)

    def team_name(self) -> str:
        return self.raw.get('team_name')

    def username(self) -> str:
        if service_user := self.service_user():
            return service_user.username()

    def password(self) -> str:
        if service_user := self.service_user():
            return service_user.passwd()

    def role(self) -> str:
        return self.raw.get('role')

    def github_auth_teams(self) -> list[str]:
        """
        returns the github auth teams: [org:team, ...]
        """
        return self.raw.get('github_auth_teams', [])

    def has_basic_auth_credentials(self) -> bool:
        if self.username() or self.password():
            return True
        return False

    def _required_attributes(self):
        return [
            'concourse_endpoint_name',
            'github_auth_teams',
            'role',
            'service_user',
            'team_name',
        ]


def get_team_cfg_by_name(cc_team_cfgs, name: str) -> typing.Optional[ConcourseTeamConfig]:
    for cc_team_cfg in cc_team_cfgs:
        if cc_team_cfg.team_name() == name:
            return cc_team_cfg


def get_main_team_cfg(cc_team_cfgs):
    return get_team_cfg_by_name(cc_team_cfgs, 'main')


class PipelineCleanupPolicy(Enum):
    CLEANUP_EXTRA_PIPELINES = 'cleanup_extra_pipelines'
    NO_CLEANUP = 'no_cleanup'


class JobMappingSet(NamedModelElement):
    def job_mappings(self):
        return {name: JobMapping(name=name, raw_dict=raw) for name, raw in self.raw.items()}

    def job_mapping_for_repo_url(self, repo_url: str, cfg_set):
        for _, job_mapping in self.job_mappings().items():
            if job_mapping.matches_repo_url(repo_url, cfg_set):
                return job_mapping

        raise ValueError(f'no matching job mapping for {repo_url=}')

    def job_mapping_for_team_name(self, team_name: str):
        for _, job_mapping in self.job_mappings().items():
            if job_mapping.concourse_target_team() == team_name:
                return job_mapping

        raise ValueError(f'no matching job mapping for {team_name=}')

    def __iter__(self):
        return self.job_mappings().values().__iter__()

    def __getitem__(self, name: str):
        return JobMapping(name=name, raw_dict=self.raw.__getitem__(name))

    def validate(self):
        tgt_team_names = {jm.team_name() for jm in self}
        for team_name in (jm.team_name() for jm in self):
            if not team_name in tgt_team_names:
                raise ModelValidationError(f'{team_name=} must only be specified once')
            tgt_team_names.remove(team_name)


class SecretNamePattern(Enum):
    PREFIX = 'cc-'
    POSTFIX = '-config'


class SecretsRepo(ModelBase):
    def github_cfg(self):
        return self.raw.get('github_cfg')

    def org(self):
        return self.raw.get('org')

    def repo(self):
        return self.raw.get('repo')


def cfg_name_from_team(team_name):
    team_name = re.sub('(.)([A-Z][a-z]+)', r'\1-\2', team_name)
    team_name = re.sub('([a-z0-9])([A-Z])', r'\1-\2', team_name).lower()
    return f'{SecretNamePattern.PREFIX.value}{team_name}{SecretNamePattern.POSTFIX.value}'


def secret_cfg_name_for_team(team_name):
    return f'{team_name}_cfg'


def secret_name_from_team(team_name: str, key_generation: int) -> str:
    if key_generation is None:
        return cfg_name_from_team(team_name)
    else:
        return f'{cfg_name_from_team(team_name)}-{key_generation}'


class JobMapping(NamedModelElement):
    def team_name(self) -> str:
        # todo: use `name` attr for that (thus enforce unique mappings)
        return self.raw.get('concourse_target_team')

    def cleanup_policy(self) -> PipelineCleanupPolicy:
        return PipelineCleanupPolicy(
            self.raw.get('cleanup_policy', PipelineCleanupPolicy.CLEANUP_EXTRA_PIPELINES)
        )

    def replication_ctx_cfg_set(self) -> str:
        return self.raw.get('replication_ctx_cfg_set')

    def github_organisations(self):
        return [
            GithubOrganisationConfig(name, raw)
            for name, raw in self.raw.get('github_orgs', {}).items()
        ]

    def secret_cfg(self) -> str:
        return self.raw.get('secret_cfg')

    def secrets_repo(self) -> typing.Optional[SecretsRepo]:
        if secrets_repo := self.raw.get('secrets_repo'):
            return SecretsRepo(secrets_repo)

    def concourse_team_cfg_name(self) -> str:
        return self.raw.get('concourse_team_cfg_name')

    def matches_repo_url(self, repo_url, cfg_set) -> bool:
        repo_url = ci.util.urlparse(repo_url)
        org, repo = repo_url.path[1:].split('/')

        for github_org_cfg in self.github_organisations():
            if not github_org_cfg.org_name() == org:
                continue

            gh_cfg = cfg_set.github(github_org_cfg.github_cfg_name())
            if not gh_cfg.matches_repo_url(urllib.parse.urlunparse(repo_url)):
                continue

            if github_org_cfg.repository_matches(repo):
                return True

        return False

    def secrets_replication_pipeline_target_cc_team_cfg_name(self) -> typing.Optional[str]:
        '''
            when set the secrets replication pipline will be rendered into the defined team
        '''
        return self.raw.get('secrets_replication_pipeline_target_cc_team_cfg_name')

    def unpause_new_pipelines(self) -> bool:
        '''Whether newly created pipelines are to be unpaused.
        '''
        return self.raw.get('unpause_new_pipelines', True)

    def unpause_deployed_pipelines(self) -> bool:
        '''Whether deployed pipelines are to be unpaused unconditionally.
        '''
        return self.raw.get('unpause_deployed_pipelines', False)

    def expose_deployed_pipelines(self) -> bool:
        '''Whether to expose pipelines after deploying them.

        Exposed pipelines are viewable by authenticated users from other teams. Note: this does
        not grant access to buildlogs and build metadata.
        '''
        return self.raw.get('expose_pipelines', True)

    def trusted_teams(self) -> typing.Iterable[str]:
        '''Pull requests created/synchronized by members of this team will automatically have
        the required label set by the webhook-dispatcher
        '''
        return self.raw.get('trusted_teams', [])

    def compliance_reporting_repo_url(self) -> str:
        '''Target repo for compliance issues based on cfg policy violations.

        If not set, reporting will still be active, just the issue generation will be disabled.
        '''
        return self.raw.get('compliance_reporting_repo_url', None)

    def _required_attributes(self):
        return [
            'concourse_target_team',
            'replication_ctx_cfg_set',
        ]

    def _optional_attributes(self):
        return [
            'compliance_reporting_repo_url',
            'concourse_team_cfg_name',
            'expose_pipelines',
            'secret_cfg',
            'secrets_replication_pipeline_target_cc_team_cfg_name',
            'secrets_repo',
            'trusted_teams',
            'unpause_deployed_pipelines',
            'unpause_new_pipelines',
        ]

    def _defaults_dict(self):
        # XXX find out why this is not honoured
        return {
            'cleanup_policy': PipelineCleanupPolicy.CLEANUP_EXTRA_PIPELINES,
        }

    def target_secret_name(self):
        '''
            k8s secret name used for replication
        '''
        return cfg_name_from_team(self.team_name())

    def target_secret_cfg_name(self):
        '''
            name of the config in the target k8s secret
        '''
        return secret_cfg_name_for_team(self.team_name())


class GithubOrganisationConfig(NamedModelElement):
    def github_cfg_name(self):
        return self.raw.get('github_cfg')

    def org_name(self):
        return self.raw.get('github_org')

    def include_repositories(self):
        return self.raw.get('include_repositories', ())

    def exclude_repositories(self):
        return self.raw.get('exclude_repositories', ())

    def repository_matches(self, repository_name: str):
        repo_filter = reutil.re_filter(
            include_regexes=self.include_repositories(),
            exclude_regexes=self.exclude_repositories(),
        )

        return repo_filter(repository_name)



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import unittest

from concourse.model.traits.slack import SlackTrait


class SlackTraitTest(unittest.TestCase):
    def test_slack_trait_validation(self):
        channel_cfgs = [
            {
                'channel_name':'foo',
                'slack_cfg_name':'bar'
            }
        ]

        examinee = SlackTrait

        # valid invocations
        examinee(
            'my_slack_trait',
            'my_variant_name',
            raw_dict={
                'default_channel': 'foo',
                'channel_cfgs': channel_cfgs
            }
        ).validate()



import logging
import typing
import datetime
import enum

import ocm
import git
import github3.repos

import cnudie.retrieve
import cnudie.util
import gitutil
import github.util
import release_notes.model as rnm
import release_notes.utils as rnu
import version

logger = logging.getLogger(__name__)


class SpecialVersion(enum.Enum):
    HEAD = enum.auto()
    INITIAL = enum.auto()


def _list_commits_since_tag(
        repo: git.Repo,
        tag: git.TagReference,
) -> tuple[tuple[git.Commit], tuple[git.Commit]]:
    '''Return a list of between the given tag and HEAD

    :return: a tuple of commits'''
    if repo.is_ancestor(tag.commit, 'HEAD'):
        logger.info(f"Commit tagged '{tag.name}' is a direct ancestor of HEAD")
        return tuple(repo.iter_commits(f'HEAD...{tag.commit.hexsha}')), tuple()

    if (
        not (merge_commit_list := repo.merge_base('HEAD', tag))
        or not (merge_commit := merge_commit_list.pop())
    ):
        raise RuntimeError('cannot find merge base')
    return (
        tuple(repo.iter_commits(f'HEAD...{merge_commit.hexsha}')),
        tuple(repo.iter_commits(f'{merge_commit.hexsha}...{tag.commit.hexsha}'))
    )


def _get_release_note_commits_tuple_for_release(
        current_version: str,
        previous_version: str,
        git_helper: gitutil.GitHelper,
        github_repo: github3.repos.Repository,
) -> tuple[tuple[git.Commit], tuple[git.Commit]]:
    '''
    :return: a tuple of commits which should be included in the release notes
    and a tuple of commits which should not be included in the release notes
    '''
    logger.info('Fetching commits for release notes')

    current_version_tag = git_helper.repo.tag(current_version)
    logger.info(f"Found tag for current version: '{current_version_tag}'")
    previous_version_tag = git_helper.repo.tag(previous_version)
    logger.info(f"Found tag for previous version: '{previous_version_tag}'")

    # if the current version tag and the previous tag are ancestors, just
    # add the range (old method)
    previous_version_tag_commit_sha = git_helper.fetch_head(
        f'refs/tags/{previous_version_tag}'
    )
    current_tag_commit_sha = git_helper.fetch_head(f'refs/tags/{current_version_tag}')
    if git_helper.repo.is_ancestor(previous_version_tag_commit_sha, current_tag_commit_sha):
        logger.info('Previous tag is an ancestor, simple range should be enough.')
        return tuple(git_helper.repo.iter_commits(
            f'{current_tag_commit_sha}...{previous_version_tag_commit_sha}')
        ), tuple()

    # otherwise, use the new method
    # find start of previous release tag
    default_head = git_helper.fetch_head(f'refs/heads/{github_repo.default_branch}')
    if not (previous_branch_starts := git_helper.repo.merge_base(
        default_head,
        previous_version_tag_commit_sha,
    )):
        raise RuntimeError('cannot find the branch start for the previous version')

    previous_branch_start: git.Commit = previous_branch_starts.pop()
    logger.info(
        f'Previous tag not an ancestor. The branch start appears to be {previous_branch_start}'
    )

    # all commits from the branch start to the previous release tag
    # should be removed from the release notes
    filter_out_commits_range = (
        f'{previous_version_tag_commit_sha}...{previous_branch_start}'
    )
    logger.debug(f'{filter_out_commits_range=}')
    filter_out_commits = git_helper.repo.iter_commits(filter_out_commits_range)

    # all commits (and release notes!) not included in {filter_out_commits} should be added to the
    # final generated release notes
    filter_in_commits_range = f'{current_tag_commit_sha}...{previous_branch_start}'
    logger.debug(f'{filter_in_commits_range=}')
    filter_in_commits = git_helper.repo.iter_commits(filter_in_commits_range)

    return tuple(filter_in_commits), tuple(filter_out_commits)


def get_release_note_commits_tuple(
        release_note_version_range: tuple[str | SpecialVersion, str | SpecialVersion],
        git_helper:gitutil.GitHelper,
        github_repo: github3.repos.Repository,
) -> tuple[tuple[git.Commit], tuple[git.Commit]]:
    '''
    :return: a tuple of commits which should be included in the release notes
    and a tuple of commits which should not be included in the release notes
    '''

    from_version, to_version = release_note_version_range

    if from_version is SpecialVersion.INITIAL:
        logger.info('Version appears to be an initial release.')
        if to_version is SpecialVersion.HEAD:
            return tuple(git_helper.repo.iter_commits()), tuple()
        else:
            return tuple(git_helper.repo.iter_commits(git_helper.repo.tag(to_version))), tuple()

    if to_version is SpecialVersion.HEAD:
        logger.info(f"Considering all commits since {from_version}")
        return _list_commits_since_tag(
            repo=git_helper.repo,
            tag=git_helper.repo.tag(from_version),
        )

    logger.info(f'Considering commits in range {from_version}...{to_version}')

    return _get_release_note_commits_tuple_for_release(
        previous_version=from_version,
        current_version=to_version,
        git_helper=git_helper,
        github_repo=github_repo,
    )


def _determine_blocks_to_include(
    filter_in_commits: tuple[git.Commit, ...],
    filter_out_commits: tuple[git.Commit, ...],
    git_helper: gitutil.GitHelper,
    github_helper: github.util.GitHubRepositoryHelper,
) -> set[rnm.SourceBlock]:
    logger.info(
        f'Found {(commit_count := len(filter_in_commits))} relevant commits for release notes '
        f'({len(filter_out_commits)} filtered out).'
    )

    commit_processing_group_size = 200
    processing_group_min_seconds = 200

    if throttled := (commit_count > commit_processing_group_size):
        logger.warning(
            'A large amount of commits needs to be processed for this release. Processing will '
            'be throttled to avoid hitting rate/quota limits.'
        )
        quotient, remainder = divmod(commit_count, commit_processing_group_size)
        estimated_time = (
            quotient * processing_group_min_seconds
            + remainder * (processing_group_min_seconds/commit_processing_group_size)
        )
        logger.warning(
            f'Estimated processing time: {datetime.timedelta(seconds=estimated_time)!s}.'
        )
        if estimated_time > 7200: # 2h, the current timeout for draft-/release steps
            raise RuntimeError(
                'Aborting release-note creation as it will not complete before reaching the '
                'timeout of two hours. Please check whether the number of commits to be scanned '
                'for this release is intentional.'
            )

    # find associated pull requests for commits
    commit_pulls = rnu.request_pull_requests_from_api(
        git_helper=git_helper,
        gh=github_helper.github,
        owner=github_helper.owner,
        repo_name=github_helper.repository_name,
        commits=[*filter_in_commits, *filter_out_commits],
        group_size=commit_processing_group_size,
        min_seconds_per_group=processing_group_min_seconds,
    )
    if throttled:
        logger.info('Finished throttled processing.')
    if commit_pulls:
        logger.info(f'Found {len(commit_pulls)} commits with associated pull requests.')
        for sha, pr_list in commit_pulls.items():
            logger.info(f"\t{sha:.6} -> {','.join(str(pr.number) for pr in pr_list)}")

    source_blocks_to_be_included: set[rnm.SourceBlock] = set()
    for filter_in_commit in filter_in_commits:
        source_blocks_to_be_included.update(rnm.iter_source_blocks(
            source=filter_in_commit,
            content=filter_in_commit.message,
        ))
        for pr in commit_pulls[filter_in_commit.hexsha]:
            if pr.body is None:
                continue
            source_blocks_to_be_included.update(rnm.iter_source_blocks(
                source=pr,
                content=pr.body,
            ))

    logger.info(f'added {len(source_blocks_to_be_included)} source blocks')

    # contains release notes which should be filtered out
    blacklisted_source_blocks: set[rnm.SourceBlock] = set()
    for filter_out_commit in filter_out_commits:
        blacklisted_source_blocks.update(rnm.iter_source_blocks(
            source=filter_out_commit,
            content=filter_out_commit.message,
        ))
        for pr in commit_pulls[filter_out_commit.hexsha]:
            if pr.body is None:
                continue
            blacklisted_source_blocks.update(rnm.iter_source_blocks(
                source=pr,
                content=pr.body,
            ))

    if blacklisted_source_blocks:
        logger.info(f'added {len(blacklisted_source_blocks)} blacklisted source blocks')

        source_blocks_to_be_included -= blacklisted_source_blocks

        logger.info(
            f'Got {len(source_blocks_to_be_included)} source blocks to consider after '
            'removing duplicates.'
        )

    return source_blocks_to_be_included


def fetch_draft_release_notes(
    current_version: str,
    component: ocm.Component,
    component_descriptor_lookup: cnudie.retrieve.ComponentDescriptorLookupById,
    version_lookup: cnudie.retrieve.VersionLookupByComponent,
    repo_path: str,
):
    known_versions: list[str] = list(version_lookup(component.identity()))

    previous_version = version.greatest_version_before(
        reference_version=current_version,
        versions=known_versions,
        ignore_prerelease_versions=True,
    ) or SpecialVersion.INITIAL

    release_note_version_range = (previous_version, SpecialVersion.HEAD)

    logger.info(
        f'Creating draft-release notes from {previous_version} to current HEAD'
    )

    source = cnudie.util.determine_main_source_for_component(component)
    github_helper = rnu.github_helper_from_github_access(source.access)
    git_helper = rnu.git_helper_from_github_access(source.access, repo_path)

    # make sure _all_ tags are available locally
    git_helper.fetch_tags()

    github_repo: github3.repos.Repository = github_helper.github.repository(
        owner=github_helper.owner,
        repository=github_helper.repository_name,
    )

    # fetch commits for release
    filter_in_commits, filter_out_commits = get_release_note_commits_tuple(
        release_note_version_range=release_note_version_range,
        git_helper=git_helper,
        github_repo=github_repo
    )

    release_note_blocks = _determine_blocks_to_include(
        filter_in_commits=filter_in_commits,
        filter_out_commits=filter_out_commits,
        git_helper=git_helper,
        github_helper=github_helper,
    )

    release_notes: set[rnm.ReleaseNote] = {
        rnm.create_release_notes_obj(
            component_descriptor_lookup=component_descriptor_lookup,
            version_lookup=version_lookup,
            source_block=source_block,
            source_component=component,
            current_component=component,
        ) for source_block in release_note_blocks
    }

    return release_notes


def fetch_release_notes(
    component: ocm.Component,
    component_descriptor_lookup: cnudie.retrieve.ComponentDescriptorLookupById,
    version_lookup: cnudie.retrieve.VersionLookupByComponent,
    repo_path: str,
    current_version: typing.Optional[str] = None,
    previous_version: typing.Optional[str] = None,
) -> set[rnm.ReleaseNote]:
    ''' Fetches and returns a set of release notes for the specified component.

    :param component: An instance of the Component class from the GCI component model.
    :param repo_path: The (local) path to the git-repository.
    :param current_version: Optional argument to retrieve release notes up to a specific version.
        If not given, the current `HEAD` is used.
    :param previous_version: Optional argument to retrieve release notes starting at a specific \
        version. If not given, the closest version to `current_version` is used.

    :return: A set of ReleaseNote objects for the specified component.
    '''

    # sanity-checks / validation
    if current_version and previous_version:
        current_semver = version.parse_to_semver(current_version)
        previous_semver = version.parse_to_semver(previous_version)

        if current_semver < previous_semver:
            logger.info(
                f'{current_version=} is a predecessor to {previous_version=}. '
                'will not generate release-notes.'
            )
            return set()

        if current_semver == previous_semver:
            logger.info(
                f'Current and previous versions given are equal ({current_version!s}), '
                'will not generate release-notes.'
            )
            return set()

    known_versions: list[str] = list(version_lookup(component.identity()))

    if not previous_version:
        if current_version:
            # if we have a current version, try to find closest match and use it
            previous_version = version.greatest_version_before(
                reference_version=current_version,
                versions=known_versions,
                ignore_prerelease_versions=True,
            )
        else:
            # if no current version was given, use latest version
            previous_version = version.greatest_version(
                versions=known_versions,
                ignore_prerelease_versions=True,
            )
        if not previous_version:
            # if still no previous version could be determined this is probably the first release.
            previous_version = SpecialVersion.INITIAL

    logger.info(
        f'current: {current_version=}, previous: {previous_version=},'
    )

    release_note_version_range = (
        previous_version or SpecialVersion.INITIAL,
        current_version or SpecialVersion.HEAD
    )

    source = cnudie.util.determine_main_source_for_component(component)
    github_helper = rnu.github_helper_from_github_access(source.access)
    git_helper = rnu.git_helper_from_github_access(source.access, repo_path)

    # make sure _all_ tags are available locally
    git_helper.fetch_tags()

    github_repo: github3.repos.Repository = github_helper.github.repository(
        owner=github_helper.owner,
        repository=github_helper.repository_name,
    )

    # fetch commits for release
    filter_in_commits, filter_out_commits = get_release_note_commits_tuple(
        release_note_version_range=release_note_version_range,
        git_helper=git_helper,
        github_repo=github_repo
    )

    release_note_blocks = _determine_blocks_to_include(
        filter_in_commits=filter_in_commits,
        filter_out_commits=filter_out_commits,
        git_helper=git_helper,
        github_helper=github_helper,
    )

    release_notes: set[rnm.ReleaseNote] = {
        rnm.create_release_notes_obj(
            component_descriptor_lookup=component_descriptor_lookup,
            version_lookup=version_lookup,
            source_block=source_block,
            source_component=component,
            current_component=component,
        ) for source_block in release_note_blocks
    }

    return release_notes



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from model.base import (
    BasicCredentials,
    NamedModelElement,
)


class JiraConfig(NamedModelElement):
    '''
    Not intended to be instantiated by users of this module
    '''

    def _required_attributes(self):
        return ['base_url', 'credentials']

    def credentials(self):
        return JiraCredentials(self.raw['credentials'])

    def base_url(self):
        return self.raw['base_url']


class JiraCredentials(BasicCredentials):
    '''
    Not intended to be instantiated by users of this module
    '''
    def email(self):
        return self.raw['email']

    def full_name(self):
        return self.raw['full_name']



# SPDX-FileCopyrightText: 2021 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import pytest

import ocm

import ctt.filters as filters


@pytest.fixture
def img(
    name='image_name',
    version='1.2.3',
    ref='image_ref:1.2.3',
    type=ocm.ArtefactType.OCI_IMAGE,
):
    def _img(name=name, version=version, ref=ref, type=type):
        return ocm.Resource(
            name=name,
            version=version,
            type=type,
            access=ocm.OciAccess(
                imageReference=ref,
            )
        )
    return _img


@pytest.fixture
def comp(name='a.b/c/e', version='1.2.3'):
    def _comp(name=name, version=version):
        return ocm.ComponentReference(
            name=name,
            componentName=name,
            version=version,
        )
    return _comp


def test_image_filter(img):
    examinee = filters.ImageFilter(
        include_image_refs=('image:1',),
        exclude_image_refs=('image:2', 'image3'),
        include_image_names=('in1', 'in2'),
        exclude_image_names=('in3',),
        include_artefact_types=(ocm.ArtefactType.OCI_IMAGE, 'unknown-type'),
        exclude_artefact_types=(ocm.ArtefactType.GIT, 'another-unknown-type')
    )

    image1 = img(ref='image:1', name='in1')

    assert examinee.matches(component=None, resource=image1)

    image2 = img(ref='image:1', name='another_name')

    assert not examinee.matches(component=None, resource=image2)

    artefact3 = img(ref='image:1', name='in1', type='unknown-type')
    assert examinee.matches(component=None, resource=artefact3)

    artefact4 = img(ref='image:1', name='in1', type=ocm.ArtefactType.GIT)
    assert not examinee.matches(component=None, resource=artefact4)


def test_component_filter(comp):
    examinee = filters.ComponentFilter(
        include_component_names=('x.o/f/c1', 'c2',),
        exclude_component_names=('x.y/z/c3',),
    )

    comp1 = comp(name='x.o/f/c1')

    assert examinee.matches(component=comp1, resource=None)

    comp2 = comp(name='x.y/z/c3')

    assert not examinee.matches(component=comp2, resource=None)



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import dataclasses
import datetime
import functools
import json
import logging
import os

import elasticsearch

import ci.util
import model
import model.elasticsearch

logger = logging.getLogger(__name__)


def default_client_if_available(cfg_factory=None) -> 'ElasticSearchClient | None':
    '''
    Returns elasticsearch client if a default config is found.
    '''
    if not ci.util._running_on_ci():
        return None

    if not cfg_factory:
        cfg_factory = ci.util.ctx().cfg_factory()

    if isinstance(cfg_factory, model.ConfigurationSet):
        return from_cfg(cfg_factory.elasticsearch())

    cfg_set = cfg_factory.cfg_set(ci.util.current_config_set_name())
    es_config = cfg_set.elasticsearch()

    return from_cfg(elasticsearch_cfg=es_config)


def from_cfg(
    elasticsearch_cfg:model.elasticsearch.ElasticSearchConfig
):
    return ElasticSearchClient(
        elasticsearch=_from_cfg(elasticsearch_cfg=elasticsearch_cfg)
    )


def _from_cfg(
    elasticsearch_cfg:model.elasticsearch.ElasticSearchConfig
):
    credentials = elasticsearch_cfg.credentials()
    return elasticsearch.Elasticsearch(
        elasticsearch_cfg.endpoints(),
        http_auth=(credentials.username(), credentials.passwd()),
    )


@functools.lru_cache()
def _metadata_dict():
    # XXX mv to concourse package; deduplicate with notify step
    if not ci.util._running_on_ci():
        return {}

    import concourse.util
    build = concourse.util.find_own_running_build()
    pipeline_metadata = concourse.util.get_pipeline_metadata()
    config_set = ci.util.ctx().cfg_factory().cfg_set(pipeline_metadata.current_config_set_name)
    concourse_cfg = config_set.concourse()

    meta_dict = {
      'build-id': build.id(),
      'build-name': build.build_number(),
      'build-job-name': pipeline_metadata.job_name,
      'build-team-name': pipeline_metadata.team_name,
      'build-pipeline-name': pipeline_metadata.pipeline_name,
      'atc-external-url': concourse_cfg.external_url(),
    }

    # XXX deduplicate; mv to concourse package
    meta_dict['concourse_url'] = ci.util.urljoin(
        meta_dict['atc-external-url'],
        'teams',
        meta_dict['build-team-name'],
        'pipelines',
        meta_dict['build-pipeline-name'],
        'jobs',
        meta_dict['build-job-name'],
        'builds',
        meta_dict['build-name'],
    )

    # XXX do not hard-code env variables
    meta_dict['effective_version'] = os.environ.get('EFFECTIVE_VERSION')
    meta_dict['component_name'] = os.environ.get('COMPONENT_NAME')
    meta_dict['creation_date'] = datetime.datetime.now().isoformat()

    return meta_dict


class ElasticSearchClient:
    def __init__(
        self,
        elasticsearch: elasticsearch.Elasticsearch,
    ):
        self._api = elasticsearch

    def store_document(
        self,
        index: str,
        body: dict,
        inject_metadata=True,
        *args,
        **kwargs,
    ):
        ci.util.check_type(index, str)
        ci.util.check_type(body, dict)
        if 'doc_type' in kwargs:
            raise ValueError(
                '''
                doc_type attribute has been deprecated - see:
                https://www.elastic.co/guide/en/elasticsearch/reference/6.0/removal-of-types.html
                '''
            )

        if inject_metadata and _metadata_dict():
            md = _metadata_dict()
            body['cc_meta'] = md

        try:
            return self._api.index(
                index=index,
                doc_type='_doc',
                body=body,
                *args,
                **kwargs,
            )
        except elasticsearch.exceptions.RequestError as e:
            logger.error(e)
            logger.error(e.info.get('error').get('caused_by'))
            raise e

    def store_documents(
        self,
        index: str,
        body: [dict],
        inject_metadata=True,
        *args,
        **kwargs,
    ):
        # Bulk-loading uses a special format: A json specifying index name and doc-type
        # (always _doc) followed by the actual document json. These pairs (one for each document)
        # are then converted to newline delimited json

        # The index json does not change for bulk-loading into a single index.
        index_json = json.dumps({
            'index': {
                '_index': index,
                '_type': '_doc'
            }
        })
        return self.store_bulk(
            body='\n'.join([f'{index_json}\n{json.dumps(d)}' for d in body]),
            inject_metadata=inject_metadata,
            *args,
            **kwargs,
        )

    def store_bulk(
        self,
        body: str,
        inject_metadata=True,
        *args,
        **kwargs,
    ):
        ci.util.check_type(body, str)

        if inject_metadata and _metadata_dict():
            def inject_meta(line):
                parsed = json.loads(line)
                if 'index' not in parsed:
                    parsed['cc_meta'] = md
                    return json.dumps(parsed)
                return line

            md = _metadata_dict()
            patched_body = '\n'.join([inject_meta(line) for line in body.splitlines()])
            body = patched_body

        return self._api.bulk(
            body=body,
            *args,
            **kwargs,
        )


def metric_to_es(
    es_client: ElasticSearchClient,
    metric,
    index_name: str,
):
    try:
        es_client.store_document(
            index=index_name,
            body=dataclasses.asdict(metric),
            inject_metadata=False,
        )
    except Exception:
        import traceback
        logger.warning(traceback.format_exc())
        logger.warning('could not send route request to elastic search')



import dataclasses
import hashlib
import json

import dacite

import oci.client as oc
import oci.docker as od
import oci.model as om


def v2_cfg_from_v1_manifest(
    manifest: om.OciImageManifestV1,
) -> od.DockerCfg:
    # we only need the latest cfg
    history = manifest.history[0]
    docker_cfg = history['v1Compatibility']
    docker_cfg = json.loads(docker_cfg)

    root_fs = {
        'diff_ids': [
            layer.digest for layer in manifest.layers
        ],
        'type': 'layers',
    }
    docker_cfg['rootfs'] = root_fs

    return dacite.from_dict(
        data_class=od.DockerCfg,
        data=docker_cfg,
        config=dacite.Config(
            cast=[tuple],
        ),
    )


def v1_manifest_to_v2(
    manifest: om.OciImageManifestV1,
    oci_client: oc.Client,
    tgt_image_ref: str,
) -> om.OciImageManifest:
    docker_cfg = v2_cfg_from_v1_manifest(manifest=manifest)
    docker_cfg = dataclasses.asdict(docker_cfg)
    docker_cfg = json.dumps(docker_cfg).encode('utf-8')

    cfg_digest = f'sha256:{hashlib.sha256(docker_cfg).hexdigest()}'
    cfg_leng = len(docker_cfg)

    oci_client.put_blob(
        image_reference=tgt_image_ref,
        digest=cfg_digest,
        octets_count=cfg_leng,
        data=docker_cfg,
    )

    manifest_v2 = om.OciImageManifest(
        config=om.OciBlobRef(
            digest=cfg_digest,
            mediaType='application/vnd.docker.container.image.v1+json',
            size=cfg_leng,
        ),
        layers=manifest.layers,
    )

    return manifest_v2, docker_cfg



import dataclasses
import datetime
import enum

import awesomeversion
import dacite
import dateutil.parser


def _parse_datetime_if_present(date: str):
    if not date:
        return None
    return dateutil.parser.isoparse(date)


def _parse_date_if_present(date: str):
    if not date:
        return None
    return dateutil.parser.isoparse(date).date()


@dataclasses.dataclass # TODO: deduplicate w/ modelclass in delivery-service
class GithubUser:
    username: str
    github_hostname: str


@dataclasses.dataclass(frozen=True)
class SprintDate:
    name: str
    display_name: str
    value: datetime.datetime


@dataclasses.dataclass(frozen=True) # TODO: deduplicate w/ modelclass in delivery-service/yp.py
class Sprint:
    name: str
    dates: frozenset[SprintDate]

    def find_sprint_date(
        self,
        name: str,
        absent_ok: bool = False,
    ) -> SprintDate | None:
        for sprint_date in self.dates:
            if sprint_date.name == name:
                return sprint_date

        if absent_ok:
            return None

        raise RuntimeError(f'did not find {name=} in {self=}')

    @staticmethod
    def from_dict(raw: dict):
        return dacite.from_dict(
            data_class=Sprint,
            data=raw,
            config=dacite.Config(
                type_hooks={datetime.datetime: _parse_datetime_if_present},
                cast=[frozenset],
            ),
        )


@dataclasses.dataclass(frozen=True)
class OsReleaseInfo:
    name: str
    reached_eol: bool
    greatest_version: str | None = None
    eol_date: datetime.date | None = None

    @property
    def parsed_version(self) -> awesomeversion.AwesomeVersion:
        return awesomeversion.AwesomeVersion(self.name)

    @staticmethod
    def from_dict(raw: dict):
        return dacite.from_dict(
            data_class=OsReleaseInfo,
            data=raw,
            config=dacite.Config(
                type_hooks={datetime.date | None: _parse_date_if_present},
            ),
        )


class StatusType(enum.StrEnum):
    ERROR = enum.auto()
    INFO = enum.auto()


@dataclasses.dataclass(frozen=True) # TODO: deduplicate with model-class delivery-service
class Status:
    type: StatusType
    msg: str


@dataclasses.dataclass(frozen=True)
class GitHubAuthCredentials:
    api_url: str
    auth_token: str



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import dataclasses
import hashlib
import json

import cryptography.hazmat.primitives.asymmetric.rsa as rsa
import cryptography.hazmat.primitives.serialization as crypto_serialiation
import pytest

import cosign
import model.signing_server
import oci.model


@pytest.fixture
def image_digest_tag() -> str:
    return 'sha256:a904c847d39ae82ec8859ce623ae14ccbfff36d53ce1490b43d9bf5caa47f33b'


@pytest.fixture
def image_reference(
    image_digest_tag: str,
) -> str:
    return f'eu.gcr.io/test/img@{image_digest_tag}'


@pytest.fixture
def payload_annotations() -> dict[str, str]:
    return {
        'key': 'val',
    }


@pytest.fixture
def payload_without_annotations(
    image_reference: str,
) -> bytes:
    return cosign.payload_bytes(
        image_reference=image_reference,
    )


@pytest.fixture
def payload_without_annotations_size(
    payload_without_annotations: bytes,
) -> int:
    return len(payload_without_annotations)


@pytest.fixture
def payload_without_annotations_digest(
    payload_without_annotations: bytes,
) -> str:
    return f'sha256:{hashlib.sha256(payload_without_annotations).hexdigest()}'


@pytest.fixture
def payload_with_annotations(
    image_reference: str,
    payload_annotations: dict[str, str],
) -> bytes:
    return cosign.payload_bytes(
        image_reference=image_reference,
        annotations=payload_annotations,
        overwrite_docker_reference=None,
    )


@pytest.fixture
def payload_with_annotations_size(
    payload_with_annotations: bytes,
) -> int:
    return len(payload_with_annotations)


@pytest.fixture
def payload_with_annotations_digest(
    payload_with_annotations: bytes,
) -> str:
    return f'sha256:{hashlib.sha256(payload_with_annotations).hexdigest()}'


@pytest.fixture
def cfg_blob(
    payload_without_annotations_digest: str,
) -> bytes:
    return cosign.cfg_blob_bytes(
        payload_digest=payload_without_annotations_digest,
    )


@pytest.fixture
def cfg_blob_size(
    cfg_blob: bytes,
) -> int:
    return len(cfg_blob)


@pytest.fixture
def cfg_blob_digest(
    cfg_blob: bytes,
) -> str:
    return f'sha256:{hashlib.sha256(cfg_blob).hexdigest()}'


def generate_public_key() -> str:
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=4096,
    )
    return private_key.public_key().public_bytes(
        encoding=crypto_serialiation.Encoding.PEM,
        format=crypto_serialiation.PublicFormat.SubjectPublicKeyInfo,
    ).decode('utf-8')


@pytest.fixture
def public_key_1() -> str:
    return generate_public_key()


@pytest.fixture
def public_key_2() -> str:
    return generate_public_key()


def test_json_marshaling_with_annotations(
    image_digest_tag: str,
    payload_with_annotations: bytes,
):
    expected_json = '{"critical":{"identity":{"docker-reference":null},' \
        '"image":{"docker-manifest-digest":' \
        f'"{image_digest_tag}"' \
        '},"type":"gardener.vnd/oci/cosign-signature"},"optional":{"key":"val"}}'

    assert payload_with_annotations.decode('utf-8') == expected_json


def test_json_marshaling_without_annotations(
    image_digest_tag: str,
    payload_without_annotations: bytes,
):
    expected_json = '{"critical":{"identity":{"docker-reference":"eu.gcr.io/test/img"},' \
        '"image":{"docker-manifest-digest":' \
        f'"{image_digest_tag}"' \
        '},"type":"gardener.vnd/oci/cosign-signature"},"optional":null}'

    assert payload_without_annotations.decode('utf-8') == expected_json


def test_raise_error_for_img_ref_without_digest():
    img_ref = 'eu.gcr.io/test/img:1.0.0'
    with pytest.raises(ValueError):
        cosign.payload_bytes(image_reference=img_ref)


def test_image_signing(
    payload_with_annotations_size: int,
    payload_with_annotations_digest: str,
    payload_without_annotations_size: int,
    payload_without_annotations_digest: str,
    cfg_blob_size: int,
    cfg_blob_digest: str,
    public_key_1: str,
    public_key_2: str,
):
    '''
    Tests signing of an imaginary image multiple times in a row, starting without an existing
    signature manifest and re-using the created manifest for subsequent signatures (append mode).

    Scenarios:
    1. Signing using `rsassa-pss` algorithm and `public_key_1`
    2. Re-signing using same algorithm and key but different signature (as algorithm enforces)
    3. Re-signing using `rsassa-pkcs1-v1-5` algorithm and `public_key_1`
    4. Re-signing using `rsassa-pss` algorithm again but with `public_key_2` and changed signature
    5. Re-signing using `rsassa-pss` algorithm and `public_key_2` again but with changed payload
    '''
    manifest = oci.model.OciImageManifest(
        config=oci.model.OciBlobRef(
            digest=cfg_blob_digest,
            mediaType='application/vnd.oci.image.config.v1+json',
            size=cfg_blob_size,
        ),
        mediaType='application/vnd.oci.image.manifest.v1+json',
        layers=[],
        annotations={},
    )

    signature = 'rsassa-pss-signature-dummy'
    signing_algorithm = model.signing_server.SigningAlgorithm.RSASSA_PSS

    signed_manifest = cosign.sign_manifest(
        manifest=manifest,
        payload_size=payload_with_annotations_size,
        payload_digest=payload_with_annotations_digest,
        signature=signature,
        signing_algorithm=signing_algorithm,
        public_key=public_key_1,
        on_exist=cosign.OnExist.APPEND,
    )

    assert len(signed_manifest.layers) == 1
    signature_layer = signed_manifest.layers[0]
    assert signature_layer.annotations.get(cosign._cosign_signature_annotation_name) == signature
    assert signature_layer.size == payload_with_annotations_size
    assert signature_layer.digest == payload_with_annotations_digest

    signature = 'rsassa-pss-signature-dummy-changed'

    manifest_resigned = cosign.sign_manifest(
        manifest=dataclasses.replace(signed_manifest), # copy of previous manifest
        payload_size=payload_with_annotations_size,
        payload_digest=payload_with_annotations_digest,
        signature=signature,
        signing_algorithm=signing_algorithm,
        public_key=public_key_1,
        on_exist=cosign.OnExist.APPEND,
    )

    assert signed_manifest == manifest_resigned # nothing should have changed during re-signing

    signature = 'rsassa-pkcs1-v1-5-signature-dummy'
    signing_algorithm = model.signing_server.SigningAlgorithm.RSASSA_PKCS1_V1_5

    manifest_different_algorithm = cosign.sign_manifest(
        manifest=dataclasses.replace(manifest_resigned), # copy of previous manifest
        payload_size=payload_with_annotations_size,
        payload_digest=payload_with_annotations_digest,
        signature=signature,
        signing_algorithm=signing_algorithm,
        public_key=public_key_1,
        on_exist=cosign.OnExist.APPEND,
    )

    assert len(manifest_different_algorithm.layers) == 2
    signature_layer = manifest_different_algorithm.layers[1]
    assert signature_layer.annotations.get(cosign._cosign_signature_annotation_name) == signature
    assert signature_layer.size == payload_with_annotations_size
    assert signature_layer.digest == payload_with_annotations_digest

    signature = 'rsassa-pss-signature-dummy-changed-again'
    signing_algorithm = model.signing_server.SigningAlgorithm.RSASSA_PSS

    manifest_different_key = cosign.sign_manifest(
        manifest=dataclasses.replace(manifest_different_algorithm), # copy of previous manifest
        payload_size=payload_with_annotations_size,
        payload_digest=payload_with_annotations_digest,
        signature=signature,
        signing_algorithm=signing_algorithm,
        public_key=public_key_2,
        on_exist=cosign.OnExist.APPEND,
    )

    assert len(manifest_different_key.layers) == 3
    signature_layer = manifest_different_key.layers[2]
    assert signature_layer.annotations.get(cosign._cosign_signature_annotation_name) == signature
    assert signature_layer.size == payload_with_annotations_size
    assert signature_layer.digest == payload_with_annotations_digest

    signature = 'rsassa-pss-signature-dummy-changed-payload'
    signing_algorithm = model.signing_server.SigningAlgorithm.RSASSA_PSS

    manifest_different_key = cosign.sign_manifest(
        manifest=dataclasses.replace(manifest_different_algorithm), # copy of previous manifest
        payload_size=payload_without_annotations_size,
        payload_digest=payload_without_annotations_digest,
        signature=signature,
        signing_algorithm=signing_algorithm,
        public_key=public_key_2,
        on_exist=cosign.OnExist.APPEND,
    )

    assert len(manifest_different_key.layers) == 4
    signature_layer = manifest_different_key.layers[3]
    assert signature_layer.annotations.get(cosign._cosign_signature_annotation_name) == signature
    assert signature_layer.size == payload_without_annotations_size
    assert signature_layer.digest == payload_without_annotations_digest

    json.dumps(manifest_different_key.as_dict()).encode('utf-8')



import logging
import typing

import checkmarx.model as cmx_model
import checkmarx.util
import github.compliance.model as gcm
import github.compliance.issue as gciss
import ocm

logger: logging.Logger = logging.getLogger(__name__)


def scan_result_group_collection(
    results: tuple[cmx_model.ScanResult],
    severity_threshold: str,
):
    def classification_callback(result: cmx_model.ScanResult) -> gcm.Severity:
        max_sev = checkmarx.util.greatest_severity(result)
        return checkmarx.util.checkmarx_severity_to_github_severity(max_sev)

    def findings_callback(result: cmx_model.ScanResult) -> bool:
        max_sev = checkmarx.util.greatest_severity(result)
        return max_sev and max_sev >= threshold

    threshold = cmx_model.Severity.from_str(severity_threshold)

    return gcm.ScanResultGroupCollection(
        results=tuple(results),
        issue_type=gciss._label_checkmarx,
        classification_callback=classification_callback,
        findings_callback=findings_callback,
    )


def scan_sources(
    checkmarx_cfg_name: str,
    component_descriptor: ocm.ComponentDescriptor,
    timeout_seconds: int=3600,
    team_id: str = None,
    threshold: str = 'medium',
    exclude_paths: typing.Sequence[str] = (),
    include_paths: typing.Sequence[str] = (),
    force: bool = False,
) -> cmx_model.FinishedScans:
    checkmarx_cfg = checkmarx.util.get_checkmarx_cfg(checkmarx_cfg_name)
    if not team_id:
        team_id = checkmarx_cfg.team_id()

    logger.info(f'using checkmarx team: {team_id}')

    checkmarx_client = checkmarx.util.create_checkmarx_client(checkmarx_cfg)

    scans = checkmarx.util.scan_sources(
        component_descriptor=component_descriptor,
        cx_client=checkmarx_client,
        team_id=team_id,
        exclude_paths=exclude_paths,
        include_paths=include_paths,
        force=force,
        timeout_seconds=timeout_seconds,
    )

    checkmarx.util.print_scans(
        scans=scans,
        threshold=cmx_model.Severity.from_str(threshold),
        routes=checkmarx_client.routes,
    )

    return scans



# SPDX-FileCopyrightText: 2021 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import abc
import dataclasses
import os


class ProcessorBase:
    @abc.abstractmethod
    def process(self, processing_job):
        raise NotImplementedError()


class NoOpProcessor(ProcessorBase):
    def process(self, processing_job):
        return processing_job


class FileFilter(ProcessorBase):
    def __init__(
        self,
        filter_files,
        base_dir='',
    ):
        self._remove_entries = []
        for path in filter_files:
            with open(os.path.join(base_dir, path)) as f:
                for line in f.readlines():
                    line = line.strip()
                    if not line or line.startswith('#'):
                        continue
                    else:
                        self._remove_entries.append(line)

    def process(self, processing_job):
        upload_request = dataclasses.replace(
            processing_job.upload_request,
            remove_files=tuple(self._remove_entries),
        )

        return dataclasses.replace(
            processing_job,
            upload_request=upload_request
        )



import base64
import collections.abc
import dataclasses
import datetime
import enum
import hashlib
import io
import json
import logging
import tempfile
import threading
import time

import dacite
import dateutil.parser
import requests
import requests.auth
import urllib
import urllib.parse
import www_authenticate

import oci.auth as oa
import oci.model as om
import oci.util

urljoin = oci.util.urljoin

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

oci_request_logger = logging.getLogger('oci.client.request_logger')
oci_request_logger.setLevel(logging.DEBUG)


def _append_b64_padding_if_missing(b64_str: str):
    if b64_str[-1] == '=':
        return b64_str

    if (mod4 := len(b64_str) % 4) == 2:
        return b64_str + '=' * 2
    elif mod4 == 3:
        return b64_str + '='
    elif mod4 == 0:
        return b64_str
    else:
        raise ValueError('this is a bug')


class AuthMethod(enum.Enum):
    BEARER = 'bearer'
    BASIC = 'basic'


@dataclasses.dataclass
class OauthToken:
    token: str
    scope: str
    expires_in: int = None
    issued_at: str = None

    def valid(self):
        issued_at = dateutil.parser.isoparse(self.issued_at)
        # pessimistically deduct 30s, to be on the safe side
        expiry_date = issued_at + datetime.timedelta(seconds=self.expires_in - 30)

        now = datetime.datetime.now(tz=datetime.timezone.utc)
        return now < expiry_date

    def __post_init__(self):
        if not self.issued_at:
            self.issued_at = datetime.datetime.now(tz=datetime.timezone.utc).isoformat()
        if not self.expires_in:
            # check if format seems to be jwt
            if self.token.count('.') >= 2:
                payload = self.token.split('.')[1]
                # add padding (JWT by convention has unpadded base64)
                payload = _append_b64_padding_if_missing(b64_str=payload)

                if (mod3 := len(payload) % 3) == 0:
                    pass
                elif mod3 == 1:
                    payload += '=='
                elif mod3 == 2:
                    payload += '='

                parsed = json.loads(base64.b64decode(payload.encode('utf-8')))

                exp = parsed['exp']
                iat = parsed['iat']

                self.expires_in = exp - iat
                self.issued_at = datetime.datetime.fromtimestamp(iat, tz=datetime.timezone.utc)\
                    .isoformat()
            else:
                # hard-code a value in the future since it is not given
                self.expires_in = datetime.timedelta(minutes=10).seconds


class OauthTokenCache:
    def __init__(self):
        self.tokens = {} # {scope: token}
        self.auth_methods = {} # {netloc: method}
        self._token_access_lock = threading.Lock()

    def token(self, scope: str):
        with self._token_access_lock:
            # purge expired tokens
            self.tokens = {s:t for s,t in self.tokens.items() if t.valid()}

            return self.tokens.get(scope)

    def set_token(self, token: OauthToken):
        if not token.valid():
            raise ValueError(f'token expired: {token=}')
        # TODO: we might compare remaining validity, and only replace existing tokens
        # if the new one has a later expiry date

        with self._token_access_lock:
            self.tokens[token.scope] = token

    def set_auth_method(self, image_reference: str, auth_method: AuthMethod):
        netloc = parse_image_reference(image_reference=image_reference).netloc
        self.auth_methods[netloc] = auth_method

    def auth_method(self, image_reference: str) -> AuthMethod | None:
        image_reference = om.OciImageReference.to_image_ref(image_reference)

        netloc = image_reference.netloc
        return self.auth_methods.get(netloc)


def parse_image_reference(image_reference: str):
    image_reference = oci.util.normalise_image_reference(image_reference)
    if not image_reference.startswith('https://'):
        image_reference = 'https://' + image_reference

    parsed_url = urllib.parse.urlparse(image_reference)
    return parsed_url


def _split_image_reference(image_reference: str):
    image_reference = oci.util.normalise_image_reference(image_reference)

    image_url = parse_image_reference(image_reference=image_reference)
    image_name_and_tag = image_url.path.lstrip('/')

    prefix = image_url.netloc

    if '@' in image_name_and_tag:
        image_name, image_tag = image_name_and_tag.rsplit('@', 1)
    elif ':' in image_name_and_tag:
        image_name, image_tag = image_name_and_tag.rsplit(':', 1)
    else:
        raise ValueError(f'{image_reference=} does not seem to contain a tag')

    return prefix, image_name, image_tag


def base_api_url(
    image_reference: str | om.OciImageReference,
) -> str:
    image_reference = om.OciImageReference.to_image_ref(image_reference)

    base_url = f'https://{image_reference.netloc}'

    return urljoin(base_url, 'v2') + '/'


class OciRoutes:
    def __init__(
        self,
        base_api_url_lookup: collections.abc.Callable[[str], str]=base_api_url,
    ):
        self.base_api_url_lookup = base_api_url_lookup

    def artifact_base_url(
        self,
        image_reference: str | om.OciImageReference,
    ) -> str:
        image_reference = om.OciImageReference.to_image_ref(image_reference)

        return urljoin(
            self.base_api_url_lookup(image_reference=str(image_reference)),
            image_reference.name,
        )

    def _blobs_url(self, image_reference: str) -> str:
        return urljoin(
            self.artifact_base_url(image_reference),
            'blobs',
        )

    def ls_tags_url(self, image_reference: str) -> str:
        return urljoin(
            self.artifact_base_url(image_reference),
            'tags',
            'list',
        )

    def uploads_url(self, image_reference: str) -> str:
        return urljoin(
            self._blobs_url(image_reference),
            'uploads',
        ) + '/'

    def single_post_blob_url(self, image_reference: str, digest: str) -> str:
        '''
        used for "single-post" monolithic upload (not supported e.g. by registry-1.docker.io)
        '''
        query = urllib.parse.urlencode({
            'digest': digest,
        })
        return self.uploads_url(image_reference=image_reference) + '?' + query

    def blob_url(self, image_reference: str | om.OciImageReference, digest: str):
        if isinstance(image_reference, om.OciImageReference):
            image_reference = str(image_reference)

        return urljoin(
            self._blobs_url(image_reference=image_reference),
            digest
        )

    def manifest_url(
        self,
        image_reference: str | om.OciImageReference,
        tag_preprocessing_callback: collections.abc.Callable[[str], str]=None,
    ) -> str:
        image_reference = om.OciImageReference.to_image_ref(image_reference)

        if not (tag := image_reference.tag):
            raise ValueError(f'{image_reference=} does not seem to contain a tag')

        if tag_preprocessing_callback:
            tag = tag_preprocessing_callback(tag)

        return urljoin(
            self.artifact_base_url(image_reference=image_reference),
            'manifests',
            tag,
        )


def _scope(image_reference: str | om.OciImageReference, action: str):
    image_reference = om.OciImageReference.to_image_ref(image_reference)

    image_name = image_reference.name
    # action = 'pull' # | pull,push | catalog
    scope = f'repository:{image_name}:{action}'

    return scope


class Client:
    def __init__(
        self,
        credentials_lookup: collections.abc.Callable,
        routes: OciRoutes=OciRoutes(),
        disable_tls_validation: bool=False,
        timeout_seconds: int=None,
        session: requests.Session=None,
        tag_preprocessing_callback: collections.abc.Callable[[str], str]=None,
        tag_postprocessing_callback: collections.abc.Callable[[str], str]=None,
    ):
        '''
        @param credentials_lookup <Callable>
        @param routes <OciRoutes>
        @param disable_tls_validation <bool>
        @param timeout_seconds <int>
        @param session <Session>
        @param tag_preprocessing_callback <Callable>
            callback which is instrumented _prior_ to interacting with the OCI registry, i.e. useful
            in case the tag has to be sanitised so it is accepted by the OCI registry
        @param tag_postprocessing_callback <Callable>
            callback which is instrumented _after_ interacting with the OCI registry, i.e. useful to
            revert required sanitisation of `tag_preprocessing_callback`
        '''
        self.credentials_lookup = credentials_lookup
        self.token_cache = OauthTokenCache()
        if not session:
            self.session = requests.Session()
        else:
            self.session = session
        self.routes = routes
        self.disable_tls_validation = disable_tls_validation
        self.tag_preprocessing_callback = tag_preprocessing_callback
        self.tag_postprocessing_callback = tag_postprocessing_callback

        if timeout_seconds:
            timeout_seconds = int(timeout_seconds)
        self.timeout_seconds = timeout_seconds

    def _authenticate(
        self,
        image_reference: str | om.OciImageReference,
        scope: str,
        remaining_retries: int=3,
    ):
        if isinstance(image_reference, om.OciImageReference):
            image_reference = str(image_reference)

        cached_auth_method = self.token_cache.auth_method(image_reference=image_reference)
        if cached_auth_method is AuthMethod.BASIC:
            return # basic-auth does not require any additional preliminary steps
        if cached_auth_method is AuthMethod.BEARER and self.token_cache.token(scope=scope):
            return # no re-auth required, yet

        if 'push' in scope:
            privileges = oa.Privileges.READWRITE
        elif 'pull' in scope:
            privileges = oa.Privileges.READONLY
        else:
            privileges = None

        oci_creds = self.credentials_lookup(
            image_reference=image_reference,
            privileges=privileges,
            absent_ok=True,
        )

        if not oci_creds:
            logger.debug(f'no credentials for {image_reference=} - attempting anonymous-auth')

        url = base_api_url(
            image_reference=str(image_reference),
        )

        res = self.session.get(
            url=url,
            verify=not self.disable_tls_validation,
            timeout=(31, 121),
        )

        auth_challenge = www_authenticate.parse(res.headers.get('www-authenticate'))

        # XXX HACK HACK: fallback to basic-auth if endpoints does not state what it wants
        if 'basic' in auth_challenge or not auth_challenge:
            self.token_cache.set_auth_method(
                image_reference=image_reference,
                auth_method=AuthMethod.BASIC,
            )
            return # no additional preliminary steps required for basic-auth
        elif 'bearer' in auth_challenge:
            bearer = auth_challenge['bearer']
            service = bearer.get('service')
            self.token_cache.set_auth_method(
                image_reference=image_reference,
                auth_method=AuthMethod.BEARER,
            )
        else:
            logger.warning(f'did not understand {auth_challenge=} - pbly a bug')

        bearer_dict = {'scope': scope}
        if service:
            bearer_dict['service'] = service

        realm = bearer['realm'] + '?' + urllib.parse.urlencode(bearer_dict)

        if oci_creds:
            auth = requests.auth.HTTPBasicAuth(
              username=oci_creds.username,
              password=oci_creds.password,
            )
        else:
            auth = None

        res = self.session.get(
            url=realm,
            verify=not self.disable_tls_validation,
            auth=auth,
            timeout=(31, 121),
        )

        if not res.ok:
            logger.warning(
                f'rq against {realm=} failed: {res.status_code=} {res.reason=} {res.content=}'
            )

            if res.status_code == 429 and remaining_retries > 0:
                logger.warning('quota was exceeded, will wait a minute and then retry again')
                time.sleep(60)
                return self._authenticate(
                    image_reference=image_reference,
                    scope=scope,
                    remaining_retries=remaining_retries - 1,
                )

        res.raise_for_status()

        token_dict = res.json()
        token_dict['scope'] = scope

        token = dacite.from_dict(
            data=token_dict,
            data_class=OauthToken,
        )

        self.token_cache.set_token(token)

    def _request(
        self,
        url: str,
        image_reference: str | om.OciImageReference,
        scope: str,
        method: str='GET',
        headers: dict=None,
        raise_for_status=True,
        warn_if_not_ok=True,
        remaining_retries: int=3,
        **kwargs,
    ):
        if not 'timeout' in kwargs and self.timeout_seconds:
            kwargs['timeout'] = self.timeout_seconds

        image_reference = om.OciImageReference.to_image_ref(image_reference)

        try:
            self._authenticate(
                image_reference=image_reference,
                scope=scope,
            )
        except requests.exceptions.ConnectionError as e:
            if remaining_retries == 0:
                raise

            logger.warning(f'caught ConnectionError, going to retry... ({remaining_retries=}); {e}')
            return self._request(
                url=url,
                image_reference=image_reference,
                scope=scope,
                method=method,
                headers=headers,
                raise_for_status=raise_for_status,
                warn_if_not_ok=warn_if_not_ok,
                remaining_retries=remaining_retries - 1,
                **kwargs,
            )

        headers = headers or {}
        headers['User-Agent'] = 'gardener-oci (python3; github.com/gardener/cc-utils)'
        auth_method = self.token_cache.auth_method(image_reference=image_reference)
        auth = None

        if auth_method is AuthMethod.BASIC:
            actions = scope.split(':')[-1]
            if 'push' in actions:
                privileges = oa.Privileges.READWRITE
            else:
                privileges = oa.Privileges.READONLY

            if oci_creds := self.credentials_lookup(
                image_reference=image_reference.original_image_reference,
                privileges=privileges,
                absent_ok=True,
            ):
                auth = oci_creds.username, oci_creds.password
            else:
                logger.debug(f'did not find any matching credentials for {image_reference=}')
        else:
            headers = {
              'Authorization': f'Bearer {self.token_cache.token(scope=scope).token}',
              **headers,
            }

        if self.disable_tls_validation and 'verify' in kwargs:
            kwargs['verify'] = False

        oci_request_logger.debug(
            msg=f'oci request sent {method=} {url=}',
            extra={
                'method': method,
                'url': url,
                'auth': auth,
                'headers': headers,
                **kwargs,
            },
        )

        try:
            timeout = kwargs.pop('timeout')
        except KeyError:
            timeout = (31, 121)

        try:
            res = self.session.request(
                method=method,
                url=url,
                auth=auth,
                headers=headers,
                timeout=timeout,
                **kwargs,
            )
        except requests.exceptions.ConnectionError as e:
            if remaining_retries == 0:
                raise

            logger.warning(f'caught ConnectionError, going to retry... ({remaining_retries=}); {e}')
            return self._request(
                url=url,
                image_reference=image_reference,
                scope=scope,
                method=method,
                headers=headers,
                raise_for_status=raise_for_status,
                warn_if_not_ok=warn_if_not_ok,
                remaining_retries=remaining_retries - 1,
                **kwargs,
            )

        if not res.ok and warn_if_not_ok:
            logger.warning(
                f'rq against {url=} failed {res.status_code=} {res.reason=} {method=} {res.content}'
            )

        if raise_for_status:
            if res.status_code != 404 and not res.ok:
                logger.debug(f'{url=} {res.content=} {res.headers=}')
            res.raise_for_status()

        return res

    def manifest_raw(
        self,
        image_reference: str | om.OciImageReference,
        absent_ok: bool=False,
        accept: str=None,
    ):
        image_reference = om.OciImageReference.to_image_ref(image_reference)

        scope = _scope(image_reference=image_reference, action='pull')

        # be backards-compatible, and also accept (legacy) docker-mimetype
        if not accept:
            accept = f'{om.OCI_MANIFEST_SCHEMA_V2_MIME}, {om.DOCKER_MANIFEST_SCHEMA_V2_MIME}'

        try:
            res = self._request(
                url=self.routes.manifest_url(
                    image_reference=image_reference,
                    tag_preprocessing_callback=self.tag_preprocessing_callback,
                ),
                image_reference=image_reference,
                scope=scope,
                warn_if_not_ok=not absent_ok,
                headers={
                    'Accept': accept,
                },
            )
        except requests.exceptions.HTTPError as he:
            if he.response.status_code == 404:
                if absent_ok:
                    return None
                raise om.OciImageNotFoundException(he) from he
            raise he

        return res

    def manifest(
        self,
        image_reference: str | om.OciImageReference,
        absent_ok: bool=False,
        accept: str=None,
    ) -> om.OciImageManifest | om.OciImageManifestList:
        '''
        returns the parsed OCI Manifest for the given image reference. If the optional `accept`
        argument is passed, the given value will be set as `Accept` HTTP Header when retrieving
        the manifest (defaults to
            application/vnd.oci.image.manifest.v1+json
            application/vnd.docker.distribution.manifest.v2+json
        , which requests a single Oci Image manifest, with a preference for the mimetype defined
        by OCI, and accepting docker's mimetype as a fallback)

        The following mimetype is also well-known:

            application/vnd.oci.image.manifest.v1+json

        If set, and the underlying OCI Artifact is a "multi-arch" artifact, than the returned
        value is (parsed into) a OciImageManifestList.

        see oci.model for both mimetype and model class definitions.

        Note that in case no `accept` header is set, the returned manifest type differs depending
        on the OCI Registry, if there actually is a multi-arch artifact.

        GCR is known to return a single OCI Image manifest (defaulting to GNU/Linux x86_64),
        whereas the registry backing quay.io will return a Manifest-List regardless of accept
        header.
        '''
        image_reference = om.OciImageReference.to_image_ref(image_reference)
        res = self.manifest_raw(
            image_reference=image_reference,
            absent_ok=absent_ok,
            accept=accept,
        )

        if not res and absent_ok:
            return None

        manifest_dict = res.json()

        if manifest_dict.get('mediaType') in (
            om.DOCKER_MANIFEST_LIST_MIME,
            om.OCI_IMAGE_INDEX_MIME,
        ):
            manifest = dacite.from_dict(
                data_class=om.OciImageManifestList,
                data=manifest_dict,
            )
            return manifest

        if (schema_version := int(manifest_dict['schemaVersion'])) == 1:
            manifest = dacite.from_dict(
                data_class=om.OciImageManifestV1,
                data=manifest_dict,
            )
            scope = _scope(image_reference=image_reference, action='pull')

            def fs_layer_to_oci_blob_ref(fs_layer: om.OciBlobRefV1):
                digest = fs_layer.blobSum

                res = self._request(
                    url=self.routes.blob_url(image_reference=image_reference, digest=digest),
                    image_reference=image_reference,
                    scope=scope,
                    method='HEAD',
                    stream=False,
                    timeout=None,
                )
                if (mediaType := res.headers['Content-Type']) == 'application/octet-stream':
                    # legacy container-images declare application/octet-stream as content-type,
                    # which is not supported by containerd
                    # https://github.com/containerd/containerd/pull/2456#issuecomment-406478687
                    mediaType = 'application/vnd.docker.image.rootfs.diff.tar.gzip'
                return om.OciBlobRef(
                    digest=digest,
                    mediaType=mediaType,
                    size=int(res.headers['Content-Length']),
                )

            manifest.layers = [
                fs_layer_to_oci_blob_ref(fs_layer) for fs_layer
                in manifest.fsLayers
            ]

            return manifest
        elif schema_version == 2:
            return dacite.from_dict(
                data_class=om.OciImageManifest,
                data=manifest_dict,
            )
        else:
            raise NotImplementedError(schema_version)

    def head_manifest(
        self,
        image_reference: str,
        absent_ok=False,
        accept: str=None,
    ) -> om.OciBlobRef | None:
        '''
        issues an HTTP-HEAD request for the specified oci-artifact's manifest and returns
        the thus-retrieved metadata if it exists.

        Note that the hash digest may be absent, or incorrect, as defined by the OCI
        distribution-spec.

        if `absent_ok` is truthy, `None` is returned in case the requested manifest does not
        exist; otherwise, requests.exceptions.HTTPError is raised in this case.

        To retrieve the actual manifest, use `self.manifest` or `self.manifest_raw`
        '''
        scope = _scope(image_reference=image_reference, action='pull')

        if not accept:
            accept = om.MimeTypes.single_image

        res = self._request(
            url=self.routes.manifest_url(
                image_reference=image_reference,
                tag_preprocessing_callback=self.tag_preprocessing_callback,
            ),
            image_reference=image_reference,
            method='HEAD',
            headers={
                'accept': accept,
            },
            scope=scope,
            stream=False,
            raise_for_status=not absent_ok,
            warn_if_not_ok=not absent_ok,
        )
        if not res.ok and absent_ok:
            return None

        headers = res.headers

        # XXX Docker-Content-Digest header may be absent or incorrect
        # -> it would be preferrable to retrieve the manifest and calculate the hash manually

        if size := headers.get('Content-Length'):
            size = int(size)

        return om.OciBlobRef(
            digest=headers.get('Docker-Content-Digest', None),
            mediaType=headers['Content-Type'],
            size=size,
        )

    def to_digest_hash(
        self,
        image_reference: str | om.OciImageReference,
        accept: str=None,
    ):
        image_reference = om.OciImageReference.to_image_ref(image_reference)
        if image_reference.has_digest_tag:
            return str(image_reference)

        manifest_hash_digest = hashlib.sha256(
            self.manifest_raw(
                image_reference=image_reference,
                accept=accept,
            ).content
        ).hexdigest()

        prefix = image_reference.ref_without_tag

        return f'{prefix}@sha256:{manifest_hash_digest}'

    def tags(self, image_reference: str):
        scope = _scope(image_reference=image_reference, action='pull')

        res = self._request(
            url=self.routes.ls_tags_url(image_reference=image_reference),
            image_reference=image_reference,
            scope=scope,
            method='GET'
        )

        res.raise_for_status()

        # Google-Artifact-Registry (maybe also others) will return http-200 + HTML in certain
        # error cases (e.g. if image_reference contains a pipe (|) character)
        try:
            tags = res.json()['tags']
        except json.decoder.JSONDecodeError as jde:
            if not (content_type := res.headers['Content-Type']) == 'application/json':
                jde.add_note(f'unexpected Content-Type: {content_type=}')

            jde.add_note(f'{image_reference=}')

            raise jde

        if self.tag_postprocessing_callback:
            tags = [
                self.tag_postprocessing_callback(tag)
                for tag in tags
            ]

        return tags

    def has_multiarch(self, image_reference: str) -> bool:
        res = self.head_manifest(
            image_reference=image_reference,
            absent_ok=True,
            accept=om.MimeTypes.multiarch,
        )
        if res:
            return True

        # sanity-check: at least single image must exist
        self.head_manifest(
            image_reference=image_reference,
            absent_ok=False,
            accept=om.MimeTypes.single_image,
        )
        return False

    def put_manifest(
        self,
        image_reference: str | om.OciImageReference,
        manifest: bytes,
    ):
        image_reference = om.OciImageReference.to_image_ref(image_reference)
        scope = _scope(image_reference=image_reference, action='push,pull')

        parsed = json.loads(manifest)
        content_type = parsed.get('mediaType', om.OCI_MANIFEST_SCHEMA_V2_MIME)

        logger.debug(f'manifest-mimetype: {content_type=}')

        res = self._request(
            url=self.routes.manifest_url(
                image_reference=image_reference,
                tag_preprocessing_callback=self.tag_preprocessing_callback,
            ),
            image_reference=image_reference,
            scope=scope,
            method='PUT',
            raise_for_status=False,
            headers={
                'Content-Type': content_type,
            },
            data=manifest,
        )

        if not res.ok:
            logger.warning(f'our manifest was rejected (see below for more details): {manifest=}')
        res.raise_for_status()

        return res

    def delete_manifest(
        self,
        image_reference: om.OciImageReference | str,
        purge: bool=False,
        accept: str=om.MimeTypes.prefer_multiarch,
    ):
        '''
        deletes the specified manifest. Depending on whether the passed image_reference contains
        a digest or a symbolic tag, the resulting semantics (and error cases) differs.

        If the image-reference contains a symbolic tag, the tag will be removed (a.k.a. untagged).
        The manifest will still be accessible, either through other tags, or through digest-tag.

        If the image-reference contains a digest, the manifest will be removed and no longer be
        accessible. However, this operation will fail if there are tags referencing the manifest.

        If `purge` is set to `True`, _and_ the image-reference contains a symbolic tag, then the
        manifest will be retrieved (to calculate the digest), then it will be untagged, then the
        manifest will be deleted. Note that the last operation may still fail if there are other
        tags referencing the same manifest.
        '''
        image_reference = om.OciImageReference(image_reference)
        scope = _scope(image_reference=image_reference, action='push,pull')

        if not purge or image_reference.has_digest_tag:
            if accept:
                headers = {'Accept': accept}
            else:
                headers = {}

            return self._request(
                url=self.routes.manifest_url(
                    image_reference=image_reference,
                    tag_preprocessing_callback=self.tag_preprocessing_callback,
                ),
                image_reference=image_reference,
                scope=scope,
                headers=headers,
                method='DELETE',
            )
        elif image_reference.has_symbolical_tag:
            manifest_raw = self.manifest_raw(
                image_reference=image_reference,
                accept=accept,
            )
            manifest_digest = f'sha256:{hashlib.sha256(manifest_raw.content).hexdigest()}'

            res = self.delete_manifest(
                image_reference=image_reference,
                purge=False,
                accept=accept,
            )
            res.raise_for_status()
            return self.delete_manifest(
                image_reference=f'{image_reference.ref_without_tag}@{manifest_digest}',
                purge=False,
            )
        else:
            raise RuntimeError('this case should not occur (this is a bug)')

    def delete_blob(
        self,
        image_reference: om.OciImageReference | str,
        digest: str,
    ):
        image_reference = om.OciImageReference(image_reference)
        scope = _scope(image_reference=image_reference, action='push,pull')

        res = self._request(
            url=self.routes.blob_url(image_reference=image_reference, digest=digest),
            image_reference=image_reference,
            scope=scope,
            method='DELETE',
        )
        res.raise_for_status()
        return res

    def blob(
        self,
        image_reference: str | om.OciImageReference,
        digest: str,
        stream=True,
        absent_ok=False,
    ) -> requests.models.Response:
        image_reference = om.OciImageReference(image_reference)

        scope = _scope(image_reference=image_reference, action='pull')

        res = self._request(
            url=self.routes.blob_url(image_reference=image_reference, digest=digest),
            image_reference=image_reference,
            scope=scope,
            method='GET',
            stream=stream,
            timeout=None,
            raise_for_status=False,
        )

        if absent_ok and res.status_code == requests.codes.NOT_FOUND: # noqa
            return None
        res.raise_for_status()

        return res

    def head_blob(
        self,
        image_reference: str | om.OciImageReference,
        digest: str,
        absent_ok=True,
    ):
        image_reference = om.OciImageReference(image_reference)
        scope = _scope(image_reference=image_reference, action='pull')

        res = self._request(
            url=self.routes.blob_url(
                image_reference=image_reference,
                digest=digest,
            ),
            method='HEAD',
            scope=scope,
            image_reference=image_reference,
            raise_for_status=False,
            warn_if_not_ok=not absent_ok,
        )

        if absent_ok and res.status_code == 404:
            return res

        res.raise_for_status()

        return res

    def put_blob(
        self,
        image_reference: str | om.OciImageReference,
        digest: str,
        octets_count: int,
        data: requests.models.Response | collections.abc.Generator | bytes | io.IOBase,
        max_chunk=1024 * 1024 * 1, # 1 MiB
        mimetype: str='application/octet-stream',
    ):
        '''
        uploads blob as part of an image-upload as specified in oci-distribution-spec:
        https://github.com/opencontainers/distribution-spec/blob/main/spec.md#push

        mimetype should not be set to a different value than the default. It is exposed for
        users seeking lowlevel control.
        '''
        image_reference = om.OciImageReference(image_reference)
        head_res = self.head_blob(
            image_reference=image_reference,
            digest=digest,
        )
        if head_res.ok:
            logger.debug(f'skipping blob upload {digest=} - already exists')
            return

        data_is_requests_resp = isinstance(data, requests.models.Response)
        data_is_generator = isinstance(data, collections.abc.Generator)
        data_is_filelike = hasattr(data, 'read')
        data_is_bytes = isinstance(data, bytes)

        if octets_count < max_chunk or data_is_filelike or data_is_bytes:
            if data_is_requests_resp:
                data = data.content
            elif data_is_generator:
                # at least GCR does not like chunked-uploads; if small enough, workaround this
                # and create one (not-that-big) bytes-obj
                _data = bytes()
                for chunk in data:
                    _data += chunk
                data = _data
            elif data_is_filelike:
                pass # if filelike, http.client will handle streaming for us

            return self._put_blob_single_post(
                image_reference=image_reference,
                digest=digest,
                octets_count=octets_count,
                data=data,
                mimetype=mimetype,
            )
        elif octets_count >= max_chunk and (data_is_generator or data_is_requests_resp):
            # workaround: write into temporary file, as at least GCR does not implement
            # chunked-upload, and requests will not properly work w/ all generators
            # (in particular, it will not work w/ our "fake" on)
            with tempfile.TemporaryFile() as tf:
                if data_is_generator:
                    for chunk in data:
                        tf.write(chunk)
                elif data_is_requests_resp:
                    while (chunk := data.raw.read(4096)):
                        tf.write(chunk)
                else:
                    # must only enter this codepath, if either generator or requests-response
                    raise RuntimeError('this line must not be reached')
                tf.seek(0)

                return self._put_blob_single_post(
                    image_reference=image_reference,
                    digest=digest,
                    octets_count=octets_count,
                    data=tf,
                    mimetype=mimetype,
                )
        else:
            if data_is_requests_resp:
                with data:
                  return self._put_blob_chunked(
                      image_reference=image_reference,
                      digest=digest,
                      octets_count=octets_count,
                      data_iterator=data.iter_content(chunk_size=max_chunk),
                      chunk_size=max_chunk,
                      mimetype=mimetype,
                  )
            else:
              raise NotImplementedError

    def _put_blob_chunked(
        self,
        image_reference: str | om.OciImageReference,
        digest: str,
        octets_count: int,
        data_iterator: collections.abc.Iterator[bytes],
        chunk_size: int=1024 * 1024 * 16, # 16 MiB
        mimetype='application/octect-stream',
    ):
        image_reference = om.OciImageReference(image_reference)
        scope = _scope(image_reference=image_reference, action='push,pull')
        logger.debug(f'chunked-put {chunk_size=}')

        # start uploading session
        res = self._request(
            url=self.routes.uploads_url(image_reference=image_reference),
            image_reference=image_reference,
            scope=scope,
            method='POST',
            headers={
                'content-length': '0',
            }
        )
        res.raise_for_status()

        upload_url = res.headers['location']

        octets_left = octets_count
        octets_sent = 0
        offset = 0
        sha256 = hashlib.sha256()

        while octets_left > 0:
            octets_to_send = min(octets_left, chunk_size)
            octets_left -= octets_to_send

            data = next(data_iterator)
            sha256.update(data)

            if not len(data) == octets_to_send:
                # sanity check to detect programming errors
                raise ValueError(f'{len(data)=} vs {octets_to_send=}')

            logger.debug(f'{octets_to_send=} {octets_left=} {len(data)=}')
            logger.debug(f'{octets_sent + offset}-{octets_sent + octets_to_send + offset}')

            crange_from = octets_sent
            crange_to = crange_from + len(data) - 1

            res = self._request(
                url=upload_url,
                image_reference=image_reference,
                scope=scope,
                method='PATCH',
                data=data,
                headers={
                 'Content-Length': str(octets_to_send),
                 'Content-Type': mimetype,
                 'Content-Range': f'{crange_from}-{crange_to}',
                 'Range': f'{crange_from}-{crange_to}',
                }
            )
            res.raise_for_status()

            upload_url = res.headers['location']

            octets_sent += len(data)

        sha256_digest = f'sha256:{sha256.hexdigest()}'

        # close uploading session
        query = urllib.parse.urlencode({'digest': sha256_digest})
        upload_url = res.headers['location'] + '?' + query
        res = self._request(
            url=upload_url,
            image_reference=image_reference,
            scope=scope,
            method='PUT',
            headers={
                 'Content-Length': '0',
            },
        )
        return res

    def _put_blob_single_post(
        self,
        image_reference: str | om.OciImageReference,
        digest: str,
        octets_count: int,
        data: bytes,
        mimetype: str='application/octet-stream',
    ):
        logger.debug(f'single-post {image_reference=} {octets_count=}')
        image_reference = om.OciImageReference(image_reference)
        scope = _scope(image_reference=image_reference, action='push,pull')

        # XXX according to distribution-spec, single-POST should also work - however
        # this seems not to be true for registry-1.docker.io. To keep the code simple(r),
        # always do a two-step upload; we might add a cfg-option (or maybe even discovery) for
        # using single-post uploads for registries that support it (such as GCR or artifactory)
        res = self._request(
            url=self.routes.uploads_url(
                image_reference=image_reference,
            ),
            image_reference=image_reference,
            scope=scope,
            method='POST',
        )

        upload_url = res.headers.get('Location')

        # returned url _may_ be relative
        if upload_url.startswith('/'):
            parsed_url = urllib.parse.urlparse(res.url)
            upload_url = f'{parsed_url.scheme}://{parsed_url.netloc}{upload_url}'

        if '?' in upload_url:
            prefix = '&'
        else:
            prefix = '?'

        upload_url += prefix + urllib.parse.urlencode({'digest': digest})

        res = self._request(
            url=upload_url,
            image_reference=image_reference,
            scope=scope,
            method='PUT',
            headers={
                'content-type': mimetype,
                'content-length': str(octets_count),
            },
            data=data,
            raise_for_status=False,
        )

        if res.ok and not res.status_code == 201: # spec says it MUST be 201
            # also, 202 indicates the upload actually did not succeed e.g. for "docker-hub"
            logger.warning(f'{image_reference=} {res.status_code=} {digest=} - PUT may have failed')

        res.raise_for_status()
        return res



# SPDX-FileCopyrightText: 2021 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import dataclasses
import typing

import ocm


@dataclasses.dataclass
class ContainerImageUploadRequest:
    source_ref: str
    target_ref: str
    remove_files: typing.Sequence[str] = ()
    reference_target_by_digest: bool = False
    retain_symbolic_tag: bool = False


@dataclasses.dataclass
class ProcessingJob:
    component: ocm.Component
    resource: ocm.Resource
    upload_request: ContainerImageUploadRequest
    processed_resource: ocm.Resource | None = None  # added after re-upload
    inject_ocm_coordinates_into_oci_manifest: bool = False
    extra_tags: tuple[str] = ()



import logging
import time

import github3.exceptions

logger = logging.getLogger(__name__)


def retry_and_throttle(function: callable, retries=5):
    '''
    decorator intended to be used for retrying/throttling functions issueing github-api-requests
    that will sporadically run into quota-issues. There is a hard-coded sleep of 60s between
    retries. Any other errors than github3.exceptions.ForbiddenError are ignored. After configured
    amount of retries, last exception is re-raised.
    '''
    def call_with_retry(*args, retries=retries, **kwargs):
        try:
            return function(*args, **kwargs)
        except github3.exceptions.ForbiddenError as fbe:
            if retries <= 0:
                raise

            if isinstance(fbe.message, bytes):
                message = fbe.message.decode('utf-8')
            else:
                message = fbe.message
            if not 'exceeded' in message:
                raise

            retries -= 1

            logger.warning(f'error from github: {fbe.message=} {retries=}')
            time.sleep(60 * 1) # 1m
            return call_with_retry(*args, retries=retries, **kwargs)

    return call_with_retry



from model.base import (
    NamedModelElement,
)


class BtpApplicationCertificate(NamedModelElement):
    def subject_template(self) -> str:
        '''
        certificate subject template to inject CN
        '''
        return self.raw['subject_template']

    def common_name(self) -> str:
        '''
        subject common name
        '''
        return self.raw['common_name']

    def scopes(self) -> list[str]:
        '''
        scopes for certificate
        '''
        return self.raw['scopes']

    def subject(self, cn: str) -> str:
        return self.subject_template().format(cn=cn)

    @staticmethod
    def parse_serial_no_from_common_name(cn: str) -> tuple[int, str]:
        '''
        Parses serial number and base from the common name.
        It expects a common name of the format '<serial_no>.<base>', otherwise
        a ValueError is raised.
        '''
        idx = cn.find('.')
        if idx <= 0:
            return 0, cn
        try:
            return int(cn[:idx]), cn[idx+1:]
        except ValueError:
            raise ValueError(f'unexpected cn: {cn}')

    def cert_service_binding(self):
        '''
        service binding used for authentication on certificate-service
        '''
        return self.raw['cert_service_binding']

    def auth_application_certificate(self):
        '''
        application certificate used for authentication on service SPS
        '''
        return self.raw['auth_application_certificate']

    def application_endpoint(self):
        '''
        endpoint for managing certificates
        '''
        return self.raw['application_endpoint']

    def application_id(self):
        '''
        application id
        '''
        return self.raw['application_id']

    def client_id(self):
        '''
        client id for application API
        '''
        return self.raw['client_id']

    def certificate_pem(self):
        '''
        certificate as PEM
        '''
        return self.raw['certificate_pem']

    def private_key_pem(self):
        '''
        certificate private key as PEM
        '''
        return self.raw['private_key_pem']

    def validity_in_days(self):
        '''
        certificate validity in days
        '''
        return self.raw['validity_in_days']

    def _required_attributes(self):
        return ['subject_template', 'common_name', 'validity_in_days',
                'cert_service_binding', 'auth_application_certificate',
                'application_endpoint', 'application_id', 'client_id',
                'scopes', 'certificate_pem', 'private_key_pem']



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from model.base import (
    NamedModelElement,
)


class VictoropsConfig(NamedModelElement):
    def api_id(self):
        return self.raw['api_token']

    def api_key(self):
        return self.raw['api_key']

    def team_slug(self):
        return self.raw['team_slug']

    def dod_policy_slug(self):
        return self.raw['dod_policy_slug']

    def mod_policy_slug(self):
        return self.raw['mod_policy_slug']



import collections.abc
import logging
import traceback

import ocm

import ccc.oci
import cnudie.iter
import cnudie.retrieve
import cnudie.util
import oci.client as oc
import oci.model as om
import version

logger = logging.getLogger(__name__)


def iter_componentversions_to_purge(
    component: ocm.Component | ocm.ComponentDescriptor,
    policy: version.VersionRetentionPolicies,
    oci_client: oc.Client=None,
) -> collections.abc.Generator[ocm.ComponentIdentity, None, None]:
    oci_ref = cnudie.util.oci_ref(component=component)
    if isinstance(component, ocm.ComponentDescriptor):
        component = component.component

    if not oci_client:
        oci_client = ccc.oci.oci_client()

    for v in version.versions_to_purge(
        versions=oci_client.tags(oci_ref.ref_without_tag),
        reference_version=component.version,
        policy=policy,
    ):
        yield ocm.ComponentIdentity(
            name=component.name,
            version=v,
        )


def remove_component_descriptor_and_referenced_artefacts(
    component: ocm.Component | ocm.ComponentDescriptor,
    oci_client: oc.Client=None,
    lookup: cnudie.retrieve.ComponentDescriptorLookupById=None,
    recursive: bool=False,
    on_error: str='abort', # todo: implement, e.g. patch-component-descriptor-and-abort
):
    if isinstance(component, ocm.ComponentDescriptor):
        component = component.component

    logger.info(f'will try to purge {component.name}:{component.version} including local resources')

    current_component = None
    resources_with_removal_errors = []
    if not oci_client:
        oci_client = ccc.oci.oci_client()

    for node in cnudie.iter.iter(
        component=component,
        lookup=lookup,
        recursion_depth=-1 if recursive else 0,
    ):
        # cnudie.iter.iter will return sequences of:
        # - component-node (always exactly one per component)
        # - resource-nodes (if any)
        # - source-nodes (if any)
        if isinstance(node, cnudie.iter.ComponentNode):
            if current_component: # skip for first iteration
                _remove_component_descriptor(
                    component=current_component,
                    oci_client=oci_client,
                )
            current_component = node.component
            continue

        if isinstance(node, cnudie.iter.SourceNode):
            continue # we ignore source-nodes for now

        if isinstance(node, cnudie.iter.ResourceNode):
            if not node.resource.relation is ocm.ResourceRelation.LOCAL:
                logger.debug(f'skipping non-local {node.resource.name=}')
                continue
            try:
                did_remove = _remove_resource(
                    node=node,
                    oci_client=oci_client,
                )
                if not did_remove:
                    logger.info(f'do not know how to remove {node.resource=}')
            except Exception as e:
                logger.warning(f'error while trying to remove {node.resource=} - {e=}')
                traceback.print_exc()
                resources_with_removal_errors.append(node)
                if on_error == 'abort':
                    logger.fatal('error encountered - aborting comoponent-descriptor-removal')
                    raise e
                else:
                    raise ValueError(f'unknown value {on_error=}')

    # remove final component (last component-component-descriptor would otherwise not be removed,
    # as we remove component-descriptors only after (trying to) remove referenced resources.
    if current_component:
        _remove_component_descriptor(
            component=component,
            oci_client=oci_client,
        )


def _remove_component_descriptor(
    component: ocm.Component,
    oci_client: oc.Client,
):
    oci_ref = cnudie.util.oci_ref(
        component=component,
    )

    oci_client.delete_manifest(
        image_reference=oci_ref,
        purge=True,
    )


def _remove_resource(
    node: cnudie.iter.ResourceNode,
    oci_client: oc.Client,
) -> bool:
    resource = node.resource
    if not resource.type in (ocm.ArtefactType.OCI_IMAGE, 'ociImage'):
        return False # we only support removal of oci-images for now

    if not resource.relation in (ocm.ResourceRelation.LOCAL, 'local'):
        return False # external resources can never be removed (as we do not "own" them)

    if not isinstance(resource.access, ocm.OciAccess):
        return False # similar to above: we only support removal of oci-images in oci-registries

    access: ocm.OciAccess = resource.access
    image_reference = om.OciImageReference(access.imageReference)

    manifest = oci_client.manifest(
        image_reference=image_reference,
        absent_ok=True,
        accept=om.MimeTypes.prefer_multiarch,
    )

    if not manifest:
        return True # nothing to do if image does not exist

    if image_reference.has_symbolical_tag:
        purge = True
    elif image_reference.has_digest_tag:
        purge = False # no need to "purge" if we were passed a digest-tag
    else:
        raise ValueError(f'cannot remove image w/o tag: {str(image_reference)}')

    oci_client.delete_manifest(
        image_reference=image_reference,
        purge=purge,
        accept=om.MimeTypes.prefer_multiarch,
    )

    if isinstance(manifest, om.OciImageManifest):
        return True

    if not isinstance(manifest, om.OciImageManifestList):
        raise ValueError(f'did not expect type {manifest=} {type(manifest)} - this is a bug')

    # multi-arch-case - try to guess other tags, and purge those
    manifest: om.OciImageManifestList

    def iter_platform_refs():
        repository = image_reference.ref_without_tag
        base_tag = image_reference.tag

        for submanifest in manifest.manifests:
            p = submanifest.platform
            yield f'{repository}:{base_tag}-{p.os}-{p.architecture}'

    for ref in iter_platform_refs():
        if not oci_client.head_manifest(
            image_reference=ref,
            absent_ok=True,
        ):
            logger.warning(f'did not find {ref=} - ignoring')
            continue

        oci_client.delete_manifest(
            image_reference=ref,
            purge=True,
        )

    return True



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from ensure import ensure_annotations
from urllib.parse import urljoin

import ci.util

CONCOURSE_API_SUFFIX = 'api/v1'


class ConcourseApiRoutesBase:
    '''
    Constructs concourse REST API endpoint URLs for the given concourse base URL and
    team name.

    Not intended to be used outside of this module.
    '''
    @ensure_annotations
    def __init__(self, base_url: str, team: str, api_suffix=CONCOURSE_API_SUFFIX):
        '''
        @param base_url: the concourse URL as used to access the dashboard with a web browser
        @param team: the concourse team name for which to construct URLs
        @param api_suffix: specifies api version (currently only api/v1)
        '''
        self.base_url = base_url
        self.team = team
        self.api_suffix = api_suffix

    def _api_url(self, *parts, **kwargs):
        prefix_team = kwargs.get('prefix_team', True)
        team_name = self.team
        base_url = self.team_url(team_name) if prefix_team else \
            urljoin(self.base_url, self.api_suffix)
        # preserve all parts of base url
        base_url += '/'

        return urljoin(base_url, '/'.join(parts))

    @staticmethod
    def running_build_url(concourse_base_url, pipeline_metadata, build_number):
        url_parts = [
            concourse_base_url,
            'teams',
            pipeline_metadata.team_name,
            'pipelines',
            pipeline_metadata.pipeline_name,
            'jobs',
            pipeline_metadata.job_name,
            'builds',
            build_number,
        ]
        return ci.util.urljoin(*url_parts)

    @ensure_annotations
    def team_url(self, team: str=None):
        if not team:
            team = self.team
        return self._api_url('teams', team, prefix_team=False)

    def login(self):
        return ci.util.urljoin(
            self.base_url,
            'sky',
            'token'
        )

    def pipelines(self):
        return self._api_url('pipelines')

    def order_pipelines(self):
        return self._api_url('pipelines', 'ordering')

    @ensure_annotations
    def pipeline(self, pipeline_name: str):
        return self._api_url('pipelines', pipeline_name)

    @ensure_annotations
    def pipeline_cfg(self, pipeline_name: str):
        return self._api_url('pipelines', pipeline_name, 'config')

    @ensure_annotations
    def unpause_pipeline(self, pipeline_name: str):
        return self._api_url('pipelines', pipeline_name, 'unpause')

    @ensure_annotations
    def pause_pipeline(self, pipeline_name: str):
        return self._api_url('pipelines', pipeline_name, 'pause')

    @ensure_annotations
    def expose_pipeline(self, pipeline_name: str):
        return self._api_url('pipelines', pipeline_name, 'expose')

    @ensure_annotations
    def resource_check(self, pipeline_name: str, resource_name: str):
        return self._api_url('pipelines', pipeline_name, 'resources', resource_name, 'check')

    @ensure_annotations
    def resource(self, pipeline_name: str, resource_name: str):
        return self._api_url('pipelines', pipeline_name, 'resources', resource_name)

    @ensure_annotations
    def resource_versions(self, pipeline_name: str, resource_name: str):
        return self._api_url('pipelines', pipeline_name, 'resources', resource_name, 'versions')

    @ensure_annotations
    def job_builds(self, pipeline_name: str, job_name: str):
        return self._api_url('pipelines', pipeline_name, 'jobs', job_name, 'builds')

    @ensure_annotations
    def job_build(self, pipeline_name: str, job_name: str, build_name: str):
        return self._api_url('pipelines', pipeline_name, 'jobs', job_name, 'builds', build_name)

    @ensure_annotations
    def job(self, pipeline_name: str, job_name: str):
        return self._api_url('pipelines', pipeline_name, 'jobs', job_name)

    @ensure_annotations
    def pause_job(self, pipeline_name: str, job_name: str):
        return self._api_url('pipelines', pipeline_name, 'jobs', job_name, 'pause')

    @ensure_annotations
    def unpause_job(self, pipeline_name: str, job_name: str):
        return self._api_url('pipelines', pipeline_name, 'jobs', job_name, 'unpause')

    @ensure_annotations
    def build_events(self, build_id):
        return self._api_url('builds', str(build_id), 'events', prefix_team=False)

    @ensure_annotations
    def build_plan(self, build_id):
        return self._api_url('builds', str(build_id), 'plan', prefix_team=False)

    @ensure_annotations
    def abort_build(self, build_id):
        return self._api_url('builds', str(build_id), 'abort', prefix_team=False)

    @ensure_annotations
    def list_workers(self):
        return self._api_url('workers', prefix_team=False)

    @ensure_annotations
    def prune_worker(self, worker_name: str):
        return self._api_url('workers', worker_name, 'prune', prefix_team=False)

    @ensure_annotations
    def pin_resource_version(self, pipeline_name: str, resource_name: str, resource_version_id: int):
        return self._api_url(
            'pipelines', pipeline_name,
            'resources', resource_name,
            'versions', str(resource_version_id),
            'pin'
        )

    @ensure_annotations
    def unpin_resource(self, pipeline_name: str, resource_name: str):
        return self._api_url('pipelines', pipeline_name, 'resources', resource_name, 'unpin')

    @ensure_annotations
    def pin_comment(self, pipeline_name: str, resource_name: str):
        return self._api_url('pipelines', pipeline_name, 'resources', resource_name, 'pin_comment')


class ConcourseApiRoutesV6_3_0(ConcourseApiRoutesBase):
    '''Routes for Concourse v6.3.0'''

    def login(self):
        return ci.util.urljoin(
            self.base_url,
            'sky',
            'issuer',
            'token',
        )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from model.base import (
    BasicCredentials,
    NamedModelElement,
)


class GitlabConfig(NamedModelElement):
    '''
    Not intended to be instantiated by users of this module
    '''

    def purpose_labels(self):
        return set(self.raw.get('purpose_labels', ()))

    def ssh_url(self):
        return self.raw.get('sshUrl')

    def http_url(self):
        return self.raw.get('httpUrl')

    def credentials(self):
        return GitlabCredentials(self.raw.get('technicalUser'))

    def _optional_attributes(self):
        return (
            'purpose_labels',
        )

    def _required_attributes(self):
        return [
            'sshUrl',
            'httpUrl',
            'technicalUser'
        ]

    def validate(self):
        super().validate()
        # validation of credentials implicitly happens in the constructor
        self.credentials()


class GitlabCredentials(BasicCredentials):
    '''
    Not intended to be instantiated by users of this module
    '''

    def email_address(self):
        return self.raw.get('emailAddress')

    def private_key(self):
        return self.raw.get('privateKey')

    def access_token(self):
        return self.raw.get('accessToken')

    def _required_attributes(self):
        required_attribs = set(super()._required_attributes())
        return required_attribs | set(('privateKey', 'emailAddress', 'accessToken'))



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import json
import typing

from dacite import from_dict
from dataclasses import dataclass
from ensure import ensure_annotations
from enum import Enum
from urllib.parse import urlparse

import sseclient

from ci.util import warning


class ResourceType(Enum):
    GIT = 'git'
    PULL_REQUEST = 'pull-request'
    TIME = 'time'
    DOCKER_IMAGE = 'docker-image'


class SetPipelineResult(Enum):
    UPDATED = 0
    CREATED = 1


class ModelBase:
    '''
    Base class for Concourse model classes

    Not intended to be instantiated by users of this module
    '''

    def __init__(self, raw:dict, concourse_api):
        self.api = concourse_api
        self.raw = raw


class ResourceVersion(ModelBase):
    '''
    Wraps a single result returned from concourse's `<resource>/versions` route.
    Both `metadata` and `version` adhere to a schema specific to the resource type.
    '''
    def id(self):
        return self.raw['id']

    def type(self):
        return self.raw['type']

    def version(self) -> dict:
        return self.raw['version'] # specific to resource type

    def metadata(self) -> dict:
        return self.raw['metadata'] # specific to resource type

    def enabled(self) -> bool:
        return self.raw['enabled']


class PipelineConfig:
    '''
    Wrapper around the dictionary received from invoking the concourse
    `pipelines/<pipeline>/config` REST API

    Not intended to be instantiated by users of this module
    '''
    @ensure_annotations
    def __init__(self, raw:dict, concourse_api, name:str):
        self.concourse_api = concourse_api
        self.name = name
        self.raw = raw['config']
        resources = self.raw.get('resources', None)
        if not resources:
            warning('Pipeline did not contain resource definitions: {p}'.format(p=name))
            raise ValueError()
        self.resources = [PipelineConfigResource(r, self) for r in resources]

    def jobs(self):
        return [Job(job, self) for job in self.raw.get('jobs')]

    def resources_of_types(self, types):
        return [r for r in self.resources if r.type in types]


class Pipeline:
    '''
    Wrapper around the dictionary received from invoking the concourse
    `pipelines/<pipeline>` REST API

    Not intended to be instantiated by users of this module
    '''
    def __init__(self, raw: dict):
        self.raw = raw

    def name(self) -> str:
        return self.raw['name']

    def team_name(self) -> str:
        return self.raw['team_name']

    def id(self) -> int:
        return self.raw['id']

    def is_paused(self) -> bool:
        return self.raw['paused']

    def is_public(self) -> bool:
        return self.raw['public']

    def is_archived(self) -> bool:
        return self.raw['archived']


class Job:
    '''
    Wrapper around the dictionary representing a job as part of a
    concourse.client.model.PipelineConfig

    Not intended to be instantiated by users of this module
    '''
    @ensure_annotations
    def __init__(self, raw: dict, pipeline: PipelineConfig):
        self.raw = raw
        self.name = raw['name']
        self.pipeline = pipeline
        self.concourse_api = pipeline.concourse_api

    def plan(self):
        return Plan(self.raw.get('plan'), self)

    def is_triggered_by_resource(self, resource_name: str):
        get_steps = self.plan().get_steps()
        for get_step in get_steps:
            if get_step['get'] == resource_name and get_step.get('trigger', False):
                return True
        return False


class ConcourseJob:
    '''Wrapper around the dictionary representing a job as returned when
    querying the Concourse REST API for a single Job configuration.

    Not intended to be instantiated by users of this module
    '''
    @ensure_annotations
    def __init__(self, raw: dict):
        self.raw = raw

    def is_paused(self):
        return self.raw.get('paused', False)

    def next_build(self):
        # Build class requires the Concourse-API, which we do not have here. Return the dict
        # instead
        if build_raw := self.raw.get('next_build'):
            return build_raw
        return None


class Plan:
    '''
    Wrapper around the dictionary representing a plan as part of a concourse.client.model.Job

    Not intended to be instantiated by users of this module
    '''
    def __init__(self, raw: dict, job: Job):
        self.raw = raw
        self.job = job

    def get_steps(self):
        return [step for step in self.raw if 'get' in step]


class PipelineConfigResource:
    '''
    Wrapper around the dictionary representing a resource as part of a
    concourse.client.model.PipelineConfig

    Not intended to be instantiated by users of this module
    '''
    @ensure_annotations
    def __init__(self, raw:dict, pipeline:PipelineConfig):
        self.pipeline = pipeline
        self.concourse_api = pipeline.concourse_api
        self.raw = raw
        self.type = raw['type']
        self.source = raw['source']
        self.name = raw['name']

    def pipeline_name(self):
        return self.pipeline.name

    def github_source(self):
        return GithubSource(self.source, self.concourse_api)

    def failing_to_check(self):
        return self.raw.get('failing_to_check', False)

    def __str__(self):
        return f'Concourse Resource {self.name=} {self.type=}'


class GithubSource:
    '''
    Wrapper around the source attribute of a concourse.client.model.PipelineConfigResource
    instance in the special case said resource is a "githubby" resource (either a git
    repository or a github-pull-request)

    Not intended to be instantiated by users of this module
    '''
    @ensure_annotations
    def __init__(self, raw:dict, concourse_api):
        self.concourse_api = concourse_api
        self.raw = raw
        self.uri = raw['uri']

    def team_name(self):
        return self.raw['team_name']

    def repo_path(self):
        return urlparse(self.uri).path

    def parse_organisation(self):
        path = self.repo_path()
        # hardcode assumption: first part always denotes organisation
        return path.split('/')[1]

    def parse_repository(self):
        path = self.repo_path()
        # hardcode assumption: second part always denotes organisation
        return path.split('/')[2]

    def hostname(self):
        return urlparse(self.uri).hostname

    def branch_name(self):
        return self.raw['branch']

    def access_token(self):
        return self.raw['access_token']

    def disable_ci_skip(self) -> bool:
        return self.raw.get('disable_ci_skip')


class Build(ModelBase):
    '''
    Wrapper around the dictionary representing a build.

    Not intended to be instantiated by users of this module
    '''

    def id(self):
        return int(self.raw.get('id'))

    def start_time(self):
        return int(self.raw.get('start_time'))

    def stop_time(self):
        return int(self.raw.get('end_time'))

    def build_number(self) -> str:
        return self.raw.get('name')

    def status(self):
        return BuildStatus(self.raw.get('status'))

    def plan(self):
        return self.api.build_plan(self.id())

    def events(self):
        return self.api.build_events(self.id())


class BuildPlan(ModelBase):
    def task_id(self, task_name: str):
        '''
        determines the task-id for the given task_name
        If the task_name is not unique, the task-id for the first-found task with
        the given name is returned.
        If no task with the given name is found, `None` is returned.
        '''
        plan = self.raw.get('plan')

        def find_tid(p):
            if 'task' in p:
                task = p.get('task')
                if task.get('name') == task_name:
                    return p.get('id') # end recursion

            for k, v in p.items():
                # recursively traverse plan dict
                if isinstance(v, dict):
                    task_id = find_tid(v)
                    if task_id:
                        return task_id
                if isinstance(v, list):
                    for element in v:
                        task_id = find_tid(element)
                        if task_id:
                            return task_id
        return find_tid(plan)

    def contains_version_ref(self, resource_version_ref: str):
        '''
        determines if the resource version reference is found in the build plan.
        If found, the resource version triggered this build.
        '''
        def has_version_ref(p):
            if 'ref' in p:
                ref = p.get('ref')
                if ref == resource_version_ref:
                    return True  # end recursion
            for k, v in p.items():
                if isinstance(v, dict):
                    if has_version_ref(v):
                        return True
                if isinstance(v, list):
                    for element in v:
                        if has_version_ref(element):
                            return True
            return False

        return has_version_ref(self.raw.get('plan'))


class BuildEvents:
    '''
    Wrapper around the event stream returned by concourse when querying the events for a
    certain build execution. The event stream is consumed using the `process_events`
    method.

    Not intended to be instantiated by users of this module
    '''

    def __init__(self, response, concourse_api):
        '''
        @param response: the unprocessed reponse object as returned from the request.
                         concourse will send an event stream (server-side events),
                         so we have to use an appropriate client to consume them
        '''
        self.api = concourse_api
        self.response = response

    def process_events(self, callback=None, filter_for_task_id=None, yield_cb=False):
        '''
        processes all received streaming events in a blocking manner until the
        'finish-task' event is reached, which marks the end of a build execution.

        An optional callback may be specified, which is called for each received event
        with the parsed event data (wrapped into a dictionary).

        @param callback: callable accepting exactly one positional argument
        '''
        client = sseclient.SSEClient(self.response)
        should_stop = False
        # pylint: disable=no-member
        # events attrib is added by response
        for event in client.events():
            if event is None or not event.data or len(event.data.strip()) == 0:
                return True
            parsed = json.loads(event.data)
            data = parsed.get('data')

            if not data:
                continue

            if filter_for_task_id:
                if data.get('origin') and data['origin'].get('id') == filter_for_task_id:
                    matches_task_filter = True
                else:
                    matches_task_filter = False
            else:
                matches_task_filter = True

            if matches_task_filter and parsed.get('event') == 'finish-task':
                should_stop = True # do not wait any longer as our task has finished

            if callback and matches_task_filter:
                result = callback(data)
                if result and yield_cb:
                    yield result

            # if 'finish-task' event is reached, we always want to stop
            if not should_stop and data.get('event') == 'end':
                should_stop = True

            if should_stop:
                client.close()
                return True
        # pylint: enable=no-member

    def iter_buildlog(self, task_id: str):
        '''
        returns an iterator yielding the build-log for the task identified by the given task_id.
        Task IDs may be retrieved from `BuildPlan#task_id`.
        '''
        def filter_log(log_data):
            if (
                not log_data.get('origin') or
                not log_data.get('payload') or
                log_data['origin'].get('id') != task_id
            ):
                return
            return log_data['payload']

        def stop_if_task_ended(event_data):
            if not event_data.event or event_data.event != 'finish-task':
                return False
            return True

        yield from self.process_events(
            callback=filter_log,
            filter_for_task_id=task_id,
            yield_cb=True
        )


class Worker(ModelBase):
    '''
    Wrapper around the dictionary representing a Concourse Worker.
    Not intended to be instantiated by users of this module
    '''
    def state(self):
        return self.raw['state']

    def name(self):
        return self.raw['name']


@dataclass
class PinnedGitVersion:
    ref: str


@dataclass
class PinnedPRVersion:
    pr: str
    ref: str


@dataclass
class PinnedTimeVersion:
    time: str


class PipelineResource:
    '''
    Wrapper around the dictionary representing a pipeline resource returned by Concourse API

    Not intended to be instantiated by users of this module
    '''
    @ensure_annotations
    def __init__(self, raw:dict, concourse_api):
        self.concourse_api = concourse_api
        self.raw = raw
        self.name = raw['name']

    def is_pinned(self) -> bool:
        if self.raw.get('pinned_version'):
            return True
        return False

    def type(self) -> ResourceType:
        return ResourceType(self.raw.get('type'))

    def pipeline_name(self) -> str:
        return self.raw.get('pipeline_name')

    def pinned_version(self) -> typing.Union[PinnedPRVersion, PinnedGitVersion, PinnedTimeVersion]:
        if self.is_pinned():
            pinned_version = self.raw.get('pinned_version')
            if self.type() is ResourceType.PULL_REQUEST:
                return from_dict(data_class=PinnedPRVersion, data=pinned_version)
            elif self.type() is ResourceType.GIT:
                return from_dict(data_class=PinnedGitVersion, data=pinned_version)
            elif self.type() is ResourceType.TIME:
                return from_dict(data_class=PinnedTimeVersion, data=pinned_version)
            else:
                raise NotImplementedError(f'Pinned version for type {self.type()} not implemented')

    def pin_comment(self) -> typing.Optional[str]:
        return self.raw.get('pin_comment')


class BuildStatus(Enum):
    ABORTED = "aborted"
    ERRORED = "errored"
    FAILED = "failed"
    PENDING = "pending"
    RUNNING = "started"
    SUCCEEDED = "succeeded"



import argparse
import dataclasses
import datetime
import json
import os
import sys

import ocm


try:
    import yaml
    _have_yaml = True
except ImportError:
    _have_yaml = False

try:
    import oci
    import oci.auth
    import oci.client
    _have_oci = True
except ImportError:
    _have_oci = False


if _have_yaml:
    _yaml_or_json_load = yaml.safe_load
else:
    _yaml_or_json_load = json.load


def _parse_yaml_or_json(path) -> dict | list:
    with open(path) as f:
        return _yaml_or_json_load(f)


def dump(component_descriptor: ocm.ComponentDescriptor, parsed):
    if parsed.out == '-':
        outfh = sys.stdout
    else:
        outfh = open(parsed.out, 'w')

    raw_dict = dataclasses.asdict(component_descriptor)

    if _have_yaml:
        yaml.dump(
            data=raw_dict,
            stream=outfh,
            Dumper=ocm.EnumValueYamlDumper,
        )
    else:
        json.dump(
            obj=raw_dict,
            fp=outfh,
            cls=ocm.EnumJSONEncoder,
        )
    outfh.flush()


def create(parsed):
    now_ts = datetime.datetime.now(datetime.timezone.utc).isoformat(
        timespec='seconds',
    ).removesuffix('+00:00') + 'Z'

    if parsed.ocm_repo:
        ocm_repos = [
            ocm.OciOcmRepository(baseUrl=parsed.ocm_repo),
        ]
    else:
        ocm_repos = []

    labels = []
    for label in parsed.labels:
        if os.path.exists(label):
            label = _parse_yaml_or_json(label)
        else:
            label = _yaml_or_json_load(label)

        label = ocm.Label(
            name=label['name'],
            value=label['value'],
        )
        labels.append(label)

    component_descriptor = ocm.ComponentDescriptor(
        meta=ocm.Metadata(),
        component=ocm.Component(
            name=parsed.name,
            version=parsed.version,
            repositoryContexts=ocm_repos,
            provider=parsed.provider,
            componentReferences=[],
            sources=[],
            resources=[],
            labels=labels,
            creationTime=now_ts,
        ),
        signatures=[],
    )

    dump(component_descriptor, parsed)


def append(parsed):
    raw = _parse_yaml_or_json(parsed.file)

    component = raw['component']

    if parsed.type in ('r', 'resource'):
        attr = component['resources']
    elif parsed.type in ('s', 'source'):
        attr = component['sources']

    if _have_yaml:
        obj = yaml.safe_load(sys.stdin)
    else:
        obj = json.load(sys.stdin)

    if isinstance(obj, list):
        attr.extend(obj)
    else:
        attr.append(obj)

    with open(parsed.file, 'w') as f:
        if _have_yaml:
            yaml.safe_dump(raw, f)
        else:
            json.dump(raw, f)


def upload(parsed):
    if not _have_oci:
        print('ERROR: `oci`-package is not available - cannot upload')
        exit(1)

    oci_client = oci.client.Client(
        credentials_lookup=oci.auth.docker_credentials_lookup(),
    )
    component_descriptor = ocm.ComponentDescriptor.from_dict(_parse_yaml_or_json(parsed.file))
    component = component_descriptor.component
    if parsed.ocm_repo:
        component.repositoryContexts.append(
            ocm.OciOcmRepository(baseUrl=parsed.ocm_repo),
        )

    ocm_tgt_repo = component_descriptor.component.current_ocm_repo
    if not ocm_tgt_repo:
        print('ERROR: must define ocm-repository either via --ocm-repo, or via component-descriptor')
        exit(1)

    oci_target_ref = component.current_ocm_repo.component_version_oci_ref(component)

    for artefact in component.iter_artefacts():
        access = artefact.access
        if not access:
            continue

        if not access.type is ocm.AccessType.LOCAL_BLOB:
            continue

        # upload local-blobs, if needed
        if oci_client.head_blob(
            image_reference=oci_target_ref,
            digest=access.localReference,
            absent_ok=True,
        ):
            continue # no need to upload again
        blob_path = os.path.join(
            parsed.blobs_dir,
            access.localReference.removeprefix('sha256:')
        )
        if not os.path.exists(blob_path):
            print(f'error: did not find expected blob at {blob_path=}')
            print(f'this blob was expected for {artefact=}')
            exit(1)
        with open(blob_path, 'rb') as f:
            oci_client.put_blob(
                image_reference=oci_target_ref,
                digest=access.localReference,
                octets_count=access.size,
                data=f,
            )

    import cnudie.upload
    print(f'Uploading OCM Component-Descriptor to: {oci_target_ref=}')
    cnudie.upload.upload_component_descriptor(
        component_descriptor=component_descriptor,
        oci_client=oci_client,
    )


def main():
    parser = argparse.ArgumentParser()
    maincmd_parsers = parser.add_subparsers(
        title='commands',
        required=True,
    )

    create_parser = maincmd_parsers.add_parser(
        'create',
        aliases=('c',),
        help='creates a minimal OCM Component Descriptor',
    )
    create_parser.add_argument('--name', default=None)
    create_parser.add_argument('--version', default=None)
    create_parser.add_argument('--provider', default=None)
    create_parser.add_argument('--ocm-repo', default=None)
    create_parser.add_argument('--label', dest='labels', action='append', default=[])
    create_parser.add_argument('--out', '-o', default='-')
    create_parser.set_defaults(callable=create)

    add_parser = maincmd_parsers.add_parser(
        'append',
        aliases=('a',),
        help='appends resources or sources to component-descriptor',
    )
    add_parser.add_argument('type', choices=('r', 'resource', 's', 'source'))
    add_parser.add_argument('--file', '-f', required=True)
    add_parser.set_defaults(callable=append)

    upload_parser = maincmd_parsers.add_parser(
        'upload',
        aliases=('u',),
        help='uploads a component-descriptor to an OCI-Registry',
    )
    upload_parser.add_argument('--file', '-f', required=True)
    upload_parser.add_argument(
        '--blobs-dir',
        required=False,
        help='optional path to lookup local-blobs. fnames must equal sha256-hexdigest',
    )
    upload_parser.add_argument('--ocm-repo', default=None)
    upload_parser.set_defaults(callable=upload)

    if len(sys.argv) < 2:
        parser.print_usage()
        exit(0)

    parsed = parser.parse_args()

    parsed.callable(parsed)


if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import typing

from concourse.model.base import (
  AttributeSpec,
  Trait,
  TraitTransformer,
  ModelBase,
)
from concourse.model.job import (
    JobVariant,
)

CHANNEL_CFG_ATTRS = (
    AttributeSpec.required(
        name='channel_name',
        doc='the slack channel name',
        type=str,
    ),
    AttributeSpec.required(
        name='slack_cfg_name',
        doc='slack_cfg name (see cc-config)',
        type=str,
    ),
)


class ChannelConfig(ModelBase):
    @classmethod
    def _attribute_specs(cls):
        return CHANNEL_CFG_ATTRS

    def channel_name(self):
        return self.raw.get('channel_name')

    def slack_cfg_name(self):
        return self.raw.get('slack_cfg_name')

    def _required_attributes(self):
        return {
            'channel_name',
            'slack_cfg_name',
        }


ATTRIBUTES = (
    AttributeSpec.required(
        name='channel_cfgs',
        doc='the slack channel configurations to use',
        type=typing.Union[typing.List[ChannelConfig], typing.Dict[str, ChannelConfig]],
    ),
    AttributeSpec.deprecated(
        name='default_channel',
        doc='**deprecated**',
        type=str,
    )
)


class SlackTrait(Trait):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def _attribute_specs(cls):
        return ATTRIBUTES

    def _children(self):
       return self.channel_cfgs()

    def channel_cfgs(self):
        channel_cfgs = self.raw.get('channel_cfgs')
        if isinstance(channel_cfgs, list):
            return [ChannelConfig(raw_dict=v) for v in channel_cfgs]
        else:
            return [ChannelConfig(raw_dict=v) for _, v in channel_cfgs.items()]

    def transformer(self):
        return SlackTraitTransformer()

    def validate(self):
        super().validate()


class SlackTraitTransformer(TraitTransformer):
    name = 'slack'

    def process_pipeline_args(self, pipeline_args: JobVariant):
        pass



import collections.abc
import dataclasses
import enum

import cnudie.retrieve
import dso.labels
import ocm


class NodeReferenceType(enum.StrEnum):
    COMPONENT_REFERENCE = 'componentReference'
    EXTRA_COMPONENT_REFS_LABEL = f'label:{dso.labels.ExtraComponentReferencesLabel.name}'


@dataclasses.dataclass(frozen=True)
class NodePathEntry:
    component: ocm.Component
    reftype: NodeReferenceType = NodeReferenceType.COMPONENT_REFERENCE


@dataclasses.dataclass
class Node:
    path: tuple[NodePathEntry]

    @property
    def component(self):
        return self.path[-1].component

    @property
    def component_id(self):
        return self.component.identity()


@dataclasses.dataclass
class ComponentNode(Node):
    pass


class ArtefactNode:
    '''
    mixin class intended to be used for ResourceNode and SourceNode to add some convenience:

    - artefact property (useful for iterating over mixed node-types)
    - iterable (useful for pattern-matching, e.g. c,a = node)
    '''
    @property
    def artefact(self) -> ocm.Resource | ocm.Source:
        if isinstance(self, ResourceNode):
            return self.resource
        if isinstance(self, SourceNode):
            return self.source
        raise TypeError('must be of type ResourceNode or SourceNode')

    def __iter__(
        self,
    ) -> collections.abc.Generator[ocm.Component | ocm.Resource | ocm.Source, None, None]:
        # pylint: disable=E1101
        yield self.component
        yield self.artefact


@dataclasses.dataclass
class ResourceNode(Node, ArtefactNode):
    resource: ocm.Resource


@dataclasses.dataclass
class SourceNode(Node, ArtefactNode):
    source: ocm.Source


class Filter:
    @staticmethod
    def components(node: Node):
        return isinstance(node, ComponentNode)

    @staticmethod
    def artefacts(node: Node):
        return isinstance(node, ArtefactNode)

    @staticmethod
    def resources(node: Node):
        return isinstance(node, ResourceNode)

    @staticmethod
    def sources(node: Node):
        return isinstance(node, SourceNode)


def iter(
    component: ocm.Component,
    lookup: cnudie.retrieve.ComponentDescriptorLookupById=None,
    recursion_depth: int=-1,
    prune_unique: bool=True,
    node_filter: collections.abc.Callable[[Node], bool]=None,
    ocm_repo: ocm.OcmRepository | str=None,
    ctx_repo: ocm.OcmRepository | str=None, # deprecated, use `ocm_repo` instead
    component_filter: collections.abc.Callable[[ocm.Component], bool]=None,
    reftype_filter: collections.abc.Callable[[NodeReferenceType], bool]=None,
) -> collections.abc.Generator[Node, None, None]:
    '''
    returns a generator yielding the transitive closure of nodes accessible from the given component.

    See `cnudie.retrieve` for retrieving components/component descriptors.

    @param component:    root component for iteration
    @param lookup:       used to lookup referenced components descriptors
                         (thus abstracting from retrieval method)
                         optional iff recursion_depth is set to 0
    @param recursion_depth: if set to a positive value, limit recursion for resolving component
                            dependencies; -1 will resolve w/o recursion limit, 0 will not resolve
                            component dependencies
    @param prune_unique: if true, redundant component-versions will only be traversed once
    @param node_filter:  use to filter emitted nodes (see Filter for predefined filters)
    @param ocm_repo:     optional OCM Repository to be used to override in the lookup
    @param ctx_repo:     deprecated, use `ocm_repo` instead
    @param component_filter: use to exclude components (and their references) from the iterator;
                             thereby `True` means the component should be filtered out
    @param reftype_filter: use to exclude components (and their references) from the iterator if
                           they are of a certain reference type; thereby `True` means the component
                           should be filtered out
    '''
    if not ocm_repo and ctx_repo:
        ocm_repo = ctx_repo

    if isinstance(component, ocm.ComponentDescriptor):
        component = component.component

    seen_component_ids = set()

    if not lookup and not recursion_depth == 0:
        raise ValueError('lookup is required if recusion is not disabled (recursion_depth==0)')

    # need to nest actual iterator to keep global state of seen component-IDs
    def inner_iter(
        component: ocm.Component,
        lookup: cnudie.retrieve.ComponentDescriptorLookupById,
        recursion_depth,
        path: tuple[NodePathEntry]=(),
        reftype: NodeReferenceType=NodeReferenceType.COMPONENT_REFERENCE,
    ):
        if component_filter and component_filter(component):
            return

        if reftype_filter and reftype_filter(reftype):
            return

        path = (*path, NodePathEntry(component, reftype))

        yield ComponentNode(
            path=path,
        )

        for resource in component.resources:
            yield ResourceNode(
                path=path,
                resource=resource,
            )

        for source in component.sources:
            yield SourceNode(
                path=path,
                source=source,
            )

        if recursion_depth == 0:
            return # stop resolving referenced components
        elif recursion_depth > 0:
            recursion_depth -= 1

        for cref in component.componentReferences:
            cref_id = ocm.ComponentIdentity(
                name=cref.componentName,
                version=cref.version,
            )

            if ocm_repo:
                referenced_component_descriptor = lookup(cref_id, ocm_repo)
            else:
                referenced_component_descriptor = lookup(cref_id)

            yield from inner_iter(
                component=referenced_component_descriptor.component,
                lookup=lookup,
                recursion_depth=recursion_depth,
                path=path,
            )

        if not (extra_crefs_label := component.find_label(
            name=dso.labels.ExtraComponentReferencesLabel.name,
        )):
            return

        extra_crefs_label: dso.labels.ExtraComponentReferencesLabel = dso.labels.deserialise_label(
            label=extra_crefs_label,
        )

        for extra_cref in extra_crefs_label.value:
            extra_cref_id = extra_cref.component_reference

            if ocm_repo:
                referenced_component_descriptor = lookup(extra_cref_id, ocm_repo)
            else:
                referenced_component_descriptor = lookup(extra_cref_id)

            yield from inner_iter(
                component=referenced_component_descriptor.component,
                lookup=lookup,
                recursion_depth=recursion_depth,
                path=path,
                reftype=NodeReferenceType.EXTRA_COMPONENT_REFS_LABEL,
            )

    for node in inner_iter(
        component=component,
        lookup=lookup,
        recursion_depth=recursion_depth,
        path=(),
    ):
        if node_filter and not node_filter(node):
            continue

        if prune_unique and isinstance(node, ComponentNode):
            if node.component.identity() in seen_component_ids:
                continue
            else:
                seen_component_ids.add(node.component_id)

        yield node


def iter_resources(
    component: ocm.Component,
    lookup: cnudie.retrieve.ComponentDescriptorLookupById=None,
    recursion_depth: int=-1,
    prune_unique: bool=True,
    component_filter: collections.abc.Callable[[ocm.Component], bool]=None,
    reftype_filter: collections.abc.Callable[[NodeReferenceType], bool]=None,
) -> collections.abc.Generator[ResourceNode, None, None]:
    '''
    curried version of `iter` w/ node-filter preset to yield only resource-nodes
    '''
    return iter(
        component=component,
        lookup=lookup,
        recursion_depth=recursion_depth,
        prune_unique=prune_unique,
        node_filter=Filter.resources,
        component_filter=component_filter,
        reftype_filter=reftype_filter,
    )



import logging
import time

import cachetools
import github3.github

logger = logging.getLogger(__name__)

_user_cache = cachetools.TTLCache(maxsize=512, ttl=60*60*12) # 12h


def is_user_active(
    username: str,
    github: github3.github.GitHub,
    retries: int=3,
    cache: cachetools.Cache=_user_cache,
) -> bool:
    have_cache = isinstance(cache, cachetools.Cache)

    if have_cache and (github, username) in cache:
        return cache[(github, username)]

    def store_user_status(status: bool):
        if not have_cache:
            print('no')
            return status
        cache[(github, username)] = status
        return status

    try:
        user = github.user(username)
        if user.as_dict().get('suspended_at'):
            return store_user_status(False)
        return store_user_status(True)
    except github3.exceptions.NotFoundError:
        logger.warning(f'{username=} not found')
        return store_user_status(False)
    except github3.exceptions.ForbiddenError as fbe:
        logger.warning(f'{fbe.errors=} {fbe.message=}')

        if not retries:
            raise

        logger.warning(f'received http-403 - maybe due to quota - will retry {retries=}')
        time.sleep(60)
        return is_user_active(
            username=username,
            github=github,
            retries=retries - 1
        )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import ci.util


class Routes:
    '''
    github-api-v3-routes missing from github3.py
    '''
    def __init__(self, github_api):
        self._base_url = github_api._github_url

    def _url(self, *parts):
        return ci.util.urljoin(self._base_url, *parts)

    def org(self, name: str):
        return self._url('orgs', name)

    def org_hooks(self, name: str):
        return ci.util.urljoin(self.org(name=name), 'hooks')

    def org_hook(self, name: str, id: str):
        return ci.util.urljoin(self.org_hooks(name=name), id)

    def org_hook_deliveries(self, name: str, id: str):
        return ci.util.urljoin(self.org_hook(name=name, id=id), 'deliveries')

    def org_hook_delivery_atttemps(self, name: str, hook_id: str, delivery_id: str):
        return ci.util.urljoin(
            self.org_hook_deliveries(name=name,id=hook_id),
            delivery_id,
            'attempts',
        )


def org_webhooks(github_api, org_name: str) -> list[dict]:
    '''
    returns webhooks for given github-org
    elements are dicts with attributes: {
        type,
        id,
        name,
        active,
        events,
        config,
        updated_at,
        created_at,
        url,
        ping_url,
        deliveries_url,
    }
    see: https://docs.github.com/en/rest/reference/orgs#webhooks
    '''
    routes = Routes(github_api=github_api)

    return github_api._get(routes.org_hooks(name=org_name))
